﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Data;
using UmlDiagramDesigner.Core;
using UmlDiagramDesigner.Core.Enums;
using UmlDiagramDesigner.Data;
using UmlDiagramDesigner.DTO;
using UmlDiagramDesigner.Services;
using UmlDiagramDesigner.ViewModels;
using UmlDiagramDesigner.ViewModels.DependentElements;
using UmlDiagramDesigner.ViewModels.DiagramItems;
using UmlDiagramDesigner.ViewModels.DiagramRelationships;

namespace UmlDiagramDesigner.Models
{
    public class DiagramToDataModelConverter
    {
        public const string DefaultDiagramName = "UnknownProject";
        private readonly HashSet<Class> _convertedClasses;
        private readonly HashSet<View> _convertedViews;

        public DiagramToDataModelConverter()
        {
            _convertedClasses = new HashSet<Class>();
            _convertedViews = new HashSet<View>();
        }

        public Diagram Convert(CanvasViewModel visualDiagram)
        {
            Initialize();

            var dataDiagram = new Diagram
            {
                Name = visualDiagram.Name ?? DefaultDiagramName,
                Classes = visualDiagram.Items.OfType<ClassDiagramItemViewModel>().Select(ConvertToClass).ToArray(),
                Views = visualDiagram.Items.OfType<ViewDiagramItemViewModel>().Select(ConvertToView).ToArray()
            };

            return dataDiagram;
        }

        private void Initialize()
        {
            _convertedClasses.Clear();
            _convertedViews.Clear();
        }

        private View ConvertToView(ViewDiagramItemViewModel diagramView)
        {
            var converted = _convertedViews.FirstOrDefault(view => view.Name == diagramView.Name);
            if (converted != null)
            {
                return converted;
            }

            var result = new View
            {
                Name = diagramView.Name,
                Attributes = diagramView.Attributes.Where(att => att.IsSelected).Select(ConvertToAttribute).ToArray()
            };

            _convertedViews.Add(result);
            return result;
        }

        private DataAttribute ConvertToAttribute(AttributeBySourceDTO attributeBySource)
        {
            var result = new DataAttribute
            {
                Name = attributeBySource.Attribute.Name,
                PartOf = ConvertToClass(attributeBySource.Source),
                DataType = attributeBySource.Attribute.DataType,
                Constraints = new HashSet<Constraint>(ExtractConstraints(attributeBySource.Attribute))
            };
            return result;
        }

        private DataAttribute ConvertToAttribute(AttributeViewModel attributeViewModel, Class source)
        {
            var result = new DataAttribute
            {
                Name = attributeViewModel.Name,
                PartOf = source,
                DataType = attributeViewModel.DataType,
                Constraints = new HashSet<Constraint>(ExtractConstraints(attributeViewModel))
            };
            return result;            
        }

        private Class ConvertToClass(ClassDiagramItemViewModel classViewModel)
        {
            //only name needs to be checked, there cannot be duplicates
            var converted = _convertedClasses.FirstOrDefault(@class => @class.Name == classViewModel.Name);
            if (converted != null) //we already converted this class, so just reuse it
            {
                return converted;
            }

            var result = new Class {Name = classViewModel.Name};
            _convertedClasses.Add(result);

            result.PrimaryKey = ConvertToPrimaryKey(classViewModel.Attributes.FirstOrDefault(att => att.IsPrimaryKey),result);

            var inheritance = classViewModel.GetOutgoingRelationships<InheritanceViewModel>().FirstOrDefault();
            if (inheritance != null)
            {
                result.Ancestor = DetermineInheritedClass(inheritance);
            }

            var children = classViewModel.GetIncomingRelationships<InheritanceViewModel>().ToArray();
            if (children.Any())
            {
                result.Children = GetChildClasses(children);
            }

            result.Attributes =
                classViewModel.Attributes.Where(att => !att.IsForeignKey && !att.IsPrimaryKey)
                    .Select(att => ConvertToAttribute(att, result))
                    .ToArray();

            result.Associations =
                classViewModel.GetOutgoingRelationships<AssociationViewModel>().Select(rel => ExtractForeignKey(rel,result)).ToArray();
            result.Aggregations =
                classViewModel.GetOutgoingRelationships<AggregationViewModel>().Select(rel => ExtractForeignKey(rel,result)).ToArray();
            result.Compositions =
                classViewModel.GetOutgoingRelationships<CompositionViewModel>().Select(rel => ExtractForeignKey(rel,result)).ToArray();

            result.Triggers = classViewModel.Triggers.Select(trigger => ConvertToTrigger(trigger,result)).ToArray();

            return result;
        }

        private ICollection<Class> GetChildClasses(ICollection<InheritanceViewModel> children)
        {
            return children.Select(inheritance => inheritance.Source)
                .OfType<ClassDiagramItemViewModel>()
                .Select(ConvertToClass)
                .ToList();
        }

        private IEnumerable<Constraint> ExtractConstraints(AttributeViewModel attribute)
        {
            var result = new HashSet<Constraint>();
            if (attribute.IsAutoGenerated)
            {
                result.Add(new AutoGeneratedConstraint());
            }
            if (attribute.IsNotNull)
            {
                result.Add(new NotNullConstraint());
            }
            if (attribute.IsNullable)
            {
                result.Add(new NullableConstraint());
            }
            return result;
        }

        private PrimaryKey ConvertToPrimaryKey(AttributeViewModel attribute, Class source)
        {
            if (attribute == null)
            {
                return new PrimaryKey()
                {
                    Name = string.Format("{0}Id",source.Name),
                    DataType = DbDataType.Integer,
                    Constraints = new HashSet<Constraint>(),
                    PartOf = source
                };
            }
            if (!attribute.IsPrimaryKey)
            {
                throw new ArgumentException("AttributeViewModel parameter has to be a primary key.");
            }

            var result = new PrimaryKey
            {
                Name = attribute.Name,
                DataType = attribute.DataType,
                Constraints = new HashSet<Constraint>(ExtractConstraints(attribute)),
                PartOf = source,
            };

            return result;
        }

        private ForeignKeyConstraint ExtractForeignKey<TRelationship>(TRelationship relationship, Class source)
            where TRelationship : RelationshipViewModelBase
        {
            if (!(relationship is IForeignKeyRelationship))
            {
                throw new ArgumentException("relationship has to implement IForeignKeyRelationship.");
            }
            var targetEntity = relationship.Target as ClassDiagramItemViewModel;
            if (targetEntity == null)
            {
                throw  new ArgumentException("Target property of relationship has to be of type EntityDiagramItemViewModel");
            }
            var sourceEntity = relationship.Source as ClassDiagramItemViewModel;
            if (sourceEntity == null)
            {
                throw new ArgumentException("Source property of relationship has to be of type EntityDiagramItemViewModel");
            }
            var sourceForeignKey = ((IForeignKeyRelationship)relationship).SourceForeignKey;

            var result = new ForeignKeyConstraint
            {
                Name = sourceForeignKey.Name,
                PartOf = source,
                SourceMultiplicity = relationship.SourceMultiplicity,
                TargetMultiplicity = relationship.TargetMultiplicity,
                References =
                    ConvertToPrimaryKey(targetEntity.Attributes.FirstOrDefault(att => att.IsPrimaryKey),
                        ConvertToClass((ClassDiagramItemViewModel) relationship.Target))
            };
            result.DataType = result.References.DataType;

            return result;
        }

        private Class DetermineInheritedClass(InheritanceViewModel inheritance)
        {
            var target = inheritance.Target as ClassDiagramItemViewModel;
            if (target == null)
            {
                throw new ArgumentException("Inheritance must have a Target property of type EntityDiagramItemViewModel set.");
            }

            return ConvertToClass(target);
        }

        private Trigger ConvertToTrigger(TriggerViewModel triggerViewModel, Class source)
        {
            var result = new Trigger {Name = triggerViewModel.Name, Event = triggerViewModel.Event, OnTable = source};
            return result;
        }
    }
}
