﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Reflection;
using UmlDiagramDesigner.Core;
using UmlDiagramDesigner.Core.Enums;
using UmlDiagramDesigner.Models.Attributes;
using UmlDiagramDesigner.ViewModels;
using UmlDiagramDesigner.ViewModels.DependentElements;
using UmlDiagramDesigner.ViewModels.DiagramItems;
using UmlDiagramDesigner.ViewModels.DiagramRelationships;

namespace UmlDiagramDesigner.Models.DiagramValidation
{
    public class RelationshipVerifier : IRelationshipVerifier
    {
        private readonly IDialogService _dialogService;
        private RelationshipViewModelBase _relationship;
        private readonly EntityValidator _validator = new EntityValidator();

        public RelationshipVerifier(IDialogService dialogService)
        {
            _dialogService = dialogService;
        }

        public bool CanCreate(RelationshipViewModelBase relationship)
        {
            PrepareRelationship(relationship);

            if (!Attribute.IsDefined(relationship.Source.GetType(), typeof(AssociatableAttribute)) || 
                !Attribute.IsDefined(relationship.Target.GetType(),typeof(AssociatableAttribute)))
            {
                return CleanupRelationshipAndReturn(false);
            }

            if (relationship is TemporaryRelationship)
            {
                relationship = ((TemporaryRelationship) relationship).BaseRelationship;
            }

            if (!DoesSourceAndTargetSupportRelationship())
            {
                return CleanupRelationshipAndReturn(false);
            }

            //multiple-relationship attribute constraint was breached
            if (!IsMultipleRelationshipAttributeFulFilled())
            {
                return CleanupRelationshipAndReturn(false);
            }


            if (Attribute.IsDefined(relationship.GetType(), typeof (RequiresConstraintAttribute)))
            {
                if (!DoesSourceAndTargetCFulFillConstraint())
                {
                    return CleanupRelationshipAndReturn(false);
                }
            }

            if(Attribute.IsDefined(relationship.GetType(), typeof(ValidateDuplicateAttributesAttribute)))
            {
                if (!SourceAndTargetContainValidAttributes())
                {
                    return CleanupRelationshipAndReturn(false);
                }
            }

            var inheritance = relationship as InheritanceViewModel;
            if (inheritance != null)
            {
                return CleanupRelationshipAndReturn(CanCreateInheritance(inheritance));
            }

            return CleanupRelationshipAndReturn(true);
        }

        private bool SourceAndTargetContainValidAttributes()
        {
            var source = _relationship.Source as ClassDiagramItemViewModel;
            var target = _relationship.Target as ClassDiagramItemViewModel;

            if (source == null || target == null)
            {
                _dialogService.Display("Undefined source or target of the relationship. Creation aborted.");
                return false;
            }

            var hierarchy = source.DepthFirstTraversal<InheritanceViewModel>()
                .Union(source.DepthFirstTraversal<InheritanceViewModel>(true))
                .ToArray();

            ValidationResult result;

            if (!(result = _validator.Validate(hierarchy.OfType<ClassDiagramItemViewModel>()
                                                        .SelectMany(entity => entity.Attributes)
                                                        .ToArray())).Successful)
            {
                _dialogService.Display(result.ErrorMessage);
                return false;
            }

            if (!(result = _validator.Validate(hierarchy.OfType<ClassDiagramItemViewModel>()
                                                        .SelectMany(entity => entity.Triggers)
                                                        .ToArray())).Successful)
            {
                _dialogService.Display(result.ErrorMessage);
                return false;
            }

            return true;
        }

        private bool DoesSourceAndTargetCFulFillConstraint()
        {
            var constraint = _relationship.GetType().GetCustomAttribute<RequiresConstraintAttribute>().Constraint;

            var target = _relationship.Target as ClassDiagramItemViewModel;

            if (target == null) return false;

            if (constraint.HasFlag(SqlConstraint.PrimaryKey))
            {
                var source = _relationship.Source as ClassDiagramItemViewModel;

                var hierarchy = source.DepthFirstTraversal<InheritanceViewModel>()
                    .Union(source.DepthFirstTraversal<InheritanceViewModel>(true))
                    .ToArray(); //get the whole inheritance hierarchy

                if (!hierarchy.OfType<ClassDiagramItemViewModel>() //check if there
                    .SelectMany(entity => entity.Attributes)
                    .Any(att => att.IsPrimaryKey))
                {
                    _dialogService.Display(
                        string.Format("Cannot create {0} association, because {1} does not contain a Primary Key.",
                            _relationship, target.Name));
                    return false;
                }
            }

            return true;
        }

        private bool IsMultipleRelationshipAttributeFulFilled()
        {
            //source/target cannot contain multiple relationship of this type
            if (!Attribute.IsDefined(_relationship.GetType(), typeof (AllowMultipleAttribute)))
            {
                if (
                    _relationship.Source.GetOutgoingRelationships()
                        .Count(rel => rel.GetType() == _relationship.GetType()) > 1)
                {
                    _dialogService.Display(string.Format("Cannot apply multiple {0} associations.", _relationship));
                    return false;
                }
            }
            return true;
        }

        private bool DoesSourceAndTargetSupportRelationship()
        {
            if (_relationship is InheritanceViewModel)
            {
                if (
                    !_relationship.Source.GetType()
                        .GetCustomAttribute<AssociatableAttribute>()
                        .SupportedAssocation.HasFlag(UmlAssociation.Inheritance) ||
                    !_relationship.Target.GetType()
                        .GetCustomAttribute<AssociatableAttribute>()
                        .SupportedAssocation.HasFlag(UmlAssociation.Inheritance))
                {
                    _dialogService.Display("Chosen objects does not support inheritance.");
                    return false;
                }
            }
            else if (_relationship is AggregationViewModel)
            {
                if (
                    !_relationship.Source.GetType()
                        .GetCustomAttribute<AssociatableAttribute>()
                        .SupportedAssocation.HasFlag(UmlAssociation.Aggregation) ||
                    !_relationship.Target.GetType()
                        .GetCustomAttribute<AssociatableAttribute>()
                        .SupportedAssocation.HasFlag(UmlAssociation.Aggregation))
                {
                    _dialogService.Display("Chosen objects does not support aggregation.");
                    return false;
                }
            }
            else if (_relationship is CompositionViewModel)
            {
                if (
                    !_relationship.Source.GetType()
                        .GetCustomAttribute<AssociatableAttribute>()
                        .SupportedAssocation.HasFlag(UmlAssociation.Composition) ||
                    !_relationship.Target.GetType()
                        .GetCustomAttribute<AssociatableAttribute>()
                        .SupportedAssocation.HasFlag(UmlAssociation.Composition))
                {
                    _dialogService.Display("Chosen objects does not support composition.");
                    return false;
                }
            }
            else if (_relationship is AssociationViewModel)
            {
                if (
                    !_relationship.Source.GetType()
                        .GetCustomAttribute<AssociatableAttribute>()
                        .SupportedAssocation.HasFlag(UmlAssociation.Association) ||
                    !_relationship.Target.GetType()
                        .GetCustomAttribute<AssociatableAttribute>()
                        .SupportedAssocation.HasFlag(UmlAssociation.Association))
                {
                    _dialogService.Display("Chosen objects does not support association.");
                    return false;
                }
            }
            else if (_relationship is DependencyViewModel)
            {
                if (
                    !_relationship.Source.GetType()
                        .GetCustomAttribute<AssociatableAttribute>()
                        .SupportedAssocation.HasFlag(UmlAssociation.Dependency) ||
                    !_relationship.Target.GetType()
                        .GetCustomAttribute<AssociatableAttribute>()
                        .SupportedAssocation.HasFlag(UmlAssociation.Dependency))
                {
                    _dialogService.Display("Chosen objects does not support dependency.");
                    return false;
                }
            }
            return true;
        }

        private bool CanCreateInheritance(InheritanceViewModel inheritance)
        {
            var tarjans = new TarjansAlgorithm();
            var cycles = tarjans.DetectCycles(CanvasViewModel.Instance);

            
            if (cycles.Count == 0)
            {
                return true;
            }
            _dialogService.Display("Cannot create this association, because it would introduce cyclic inheritance.");
            return false;
        }

        private bool CleanupRelationshipAndReturn(bool result)
        {
            if (_relationship.Source.Relationships.Contains(_relationship))
            {
                _relationship.Source.Relationships.Remove(_relationship);
            }
            if (_relationship.Target.Relationships.Contains(_relationship))
            {
                _relationship.Target.Relationships.Remove(_relationship);
            }
            _relationship = null;
            return result;
        }

        private void PrepareRelationship(RelationshipViewModelBase relationship)
        {
            if (relationship is TemporaryRelationship)
            {
                var tempSource = relationship.Source;
                var tempTarget = relationship.Target;
                relationship = ((TemporaryRelationship) relationship).BaseRelationship;
                relationship.Source = tempSource;
                relationship.Target = tempTarget;
            }
            relationship.Source.Relationships.Add(relationship);
            relationship.Target.Relationships.Add(relationship);

            _relationship = relationship;
        }
    }
}
