﻿using System.Windows;
using UmlDiagramDesigner.Core.Graph;
using UmlDiagramDesigner.ViewModels.DiagramItems;


namespace UmlDiagramDesigner.Models.PathFinders
{


    /// <summary>
    /// Interface for objects, that are impelenting a certain pathfinding algorithm. Use FindPath method
    /// to find the path and don't forget to set it with SetPath method.
    /// </summary>
    public interface IPathFinder
    {
        /// <summary>
        /// Finds path between two collections of points.
        /// </summary>
        /// <param name="setA">The set a.</param>
        /// <param name="setB">The set b.</param>
        /// <returns>Array of points, where first element is startpoint, last element is end point.
        /// Everything else is a curviness of the path</returns>
        Point[] FindPath(Point[] setA, Point[] setB);

        /// <summary>
        /// Finds path from one set to another. You will most likely achieve better performace
        /// if you use this method and set <param name="possibleStraightLine"></param> to true.
        /// </summary>
        /// <param name="setA"></param>
        /// <param name="setB"></param>
        /// <param name="possibleStraightLine"></param>
        /// <returns>Array of points, where first element is startpoint, last element is end point.
        /// Everything else is a curviness of the path</returns>
        Point[] FindPath(Point[] setA, Point[] setB, bool possibleStraightLine);

        /// <summary>
        /// Finds path from Rectangle property of DiagramItemViewModelBase.
        /// </summary>
        /// <param name="item1">The item1.</param>
        /// <param name="item2">The item2.</param>
        /// <returns>Array of points, where first element is startpoint, last element is end point.
        /// Everything else is a curviness of the path</returns>
        Point[] FindPath(DiagramItemViewModelBase item1, DiagramItemViewModelBase item2);

        /// <summary>
        /// Finds path from one item to another. You will most likely achieve better performace
        /// if you use this method and set <param name="possibleStraightLine"></param> to true.
        /// </summary>
        /// <param name="item2"></param>
        /// <param name="possibleStraightLine"></param>
        /// <param name="item1"></param>
        /// <returns>Array of points, where first element is startpoint, last element is end point.
        /// Everything else is a curviness of the path</returns>
        Point[] FindPath(DiagramItemViewModelBase item1, DiagramItemViewModelBase item2, bool possibleStraightLine);

        /// <summary>
        /// Finds the path from <param name="item"></param> to <param name="point"></param>.
        /// </summary>
        /// <param name="item">The item.</param>
        /// <param name="point">The point.</param>
        /// <param name="possibleStraightLine">if set to <c>true</c> [possible straight line].</param>
        /// <returns></returns>
        Point[] FindPath(DiagramItemViewModelBase item, Point point, bool possibleStraightLine);

        /// <summary>
        /// Finds the path from <param name="item"></param> to <param name="points"></param>.
        /// </summary>
        /// <param name="item">The item.</param>
        /// <param name="points"></param>
        /// <returns></returns>
        Point[] FindPath(DiagramItemViewModelBase item, Point[] points);

        /// <summary>
        /// Sets the path to the association.
        /// </summary>
        /// <param name="points">The points.</param>
        /// <param name="association">The association.</param>
        void SetPath(Point[] points, IAssociation<DiagramItemViewModelBase> association);
    }
}
