﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using System.Windows;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Shapes;
using UmlDiagramDesigner.Core.Enums;
using UmlDiagramDesigner.Helpers;
using UmlDiagramDesigner.Properties;
using UmlDiagramDesigner.UserControls;
using UmlDiagramDesigner.ViewModels.DiagramRelationships;
using Rectangle = System.Drawing.Rectangle;

namespace UmlDiagramDesigner.Models.Shapes
{
    public abstract class RelationshipShapeBase : Shape
    {
        #region Dependency Properties

        public static readonly DependencyProperty X1Property = DependencyProperty.Register(
            "X1", typeof(int), typeof(RelationshipShapeBase), new FrameworkPropertyMetadata(default(int), FrameworkPropertyMetadataOptions.AffectsRender));

        public static readonly DependencyProperty Y1Property = DependencyProperty.Register(
            "Y1", typeof(int), typeof(RelationshipShapeBase), new FrameworkPropertyMetadata(default(int), FrameworkPropertyMetadataOptions.AffectsRender));

        public static readonly DependencyProperty X2Property = DependencyProperty.Register(
            "X2", typeof(int), typeof(RelationshipShapeBase), new FrameworkPropertyMetadata(default(int), FrameworkPropertyMetadataOptions.AffectsRender));

        public static readonly DependencyProperty Y2Property = DependencyProperty.Register(
            "Y2", typeof(int), typeof(RelationshipShapeBase), new FrameworkPropertyMetadata(default(int), FrameworkPropertyMetadataOptions.AffectsRender));

        public static readonly DependencyProperty SourceTextProperty = DependencyProperty.Register(
            "SourceText", typeof(string), typeof(RelationshipShapeBase), new FrameworkPropertyMetadata("", FrameworkPropertyMetadataOptions.AffectsRender));

        public static readonly DependencyProperty TargetTextProperty = DependencyProperty.Register(
            "TargetText", typeof(string), typeof(RelationshipShapeBase), new FrameworkPropertyMetadata("", FrameworkPropertyMetadataOptions.AffectsRender));

        public static readonly DependencyProperty MiddleTextProperty = DependencyProperty.Register(
            "MiddleText", typeof(string), typeof(RelationshipShapeBase), new FrameworkPropertyMetadata("", FrameworkPropertyMetadataOptions.AffectsRender));

        public static readonly DependencyProperty AdditionalPointsProperty = DependencyProperty.Register(
            "AdditionalPoints", typeof (IEnumerable<Point>), typeof (RelationshipShapeBase),
            new FrameworkPropertyMetadata(null, FrameworkPropertyMetadataOptions.AffectsRender));

        public static readonly DependencyProperty EndShapeSizeProperty = DependencyProperty.Register(
            "EndShapeSize", typeof (int), typeof (RelationshipShapeBase), new PropertyMetadata(13));

        public int EndShapeSize
        {
            get { return (int) GetValue(EndShapeSizeProperty); }
            set { SetValue(EndShapeSizeProperty, value); }
        }

        public List<Point> AdditionalPoints
        {
            get { return (List<Point>) GetValue(AdditionalPointsProperty); }
            set { SetValue(AdditionalPointsProperty, value); }
        }

        public string MiddleText
        {
            get { return (string)GetValue(MiddleTextProperty); }
            set { SetValue(MiddleTextProperty, value); }
        }

        public string TargetText
        {
            get { return (string)GetValue(TargetTextProperty); }
            set { SetValue(TargetTextProperty, value); }
        }

        public string SourceText
        {
            get { return (string)GetValue(SourceTextProperty); }
            set { SetValue(SourceTextProperty, value); }
        }

        public int Y2
        {
            get { return (int)GetValue(Y2Property); }
            set
            {
                SetValue(Y2Property, value);
            }
        }


        public int X2
        {
            get { return (int)GetValue(X2Property); }
            set
            {
                SetValue(X2Property, value);
            }
        }

        public int Y1
        {
            get { return (int)GetValue(Y1Property); }
            set
            {
                SetValue(Y1Property, value);
            }
        }

        public int X1
        {
            get { return (int)GetValue(X1Property); }
            set
            {
                SetValue(X1Property, value);
            }
        }

        public static readonly DependencyProperty FontSizeProperty = DependencyProperty.Register(
            "FontSize", typeof (int), typeof (RelationshipShapeBase), new PropertyMetadata(default(int)));

        public int FontSize
        {
            get { return (int) GetValue(FontSizeProperty); }
            set { SetValue(FontSizeProperty, value); }
        }

        #endregion

        protected Pen StrokePen;
        private Geometry _geometry;
        protected Point EndPoint;

        private int _hitTestTolerance = 20;

        protected RelationshipViewModelBase ViewModel;

        protected RelationshipShapeBase()
        {
            MouseEnter += OnMouseEnter;
            MouseLeave += OnMouseLeave;

            Settings.Default.PropertyChanged += (o, args) =>
            {
                if (args.PropertyName == "RelationshipStrokeThickness")
                {
                    StrokePen = new Pen(Brushes.Black, Settings.Default.RelationshipStrokeThickness);
                    StrokePen.Freeze();
                    StrokeThickness = StrokePen.Thickness;
                }
            };

            StrokePen = new Pen(Brushes.Black, Settings.Default.RelationshipStrokeThickness);
            StrokePen.Freeze();

            EndShapeSize = 20;
        }

        protected override HitTestResult HitTestCore(PointHitTestParameters hitTestParameters)
        {
            var result = base.HitTestCore(hitTestParameters); //original hit test

            if (result == null && _geometry != null) //previous hit test failed and geometry is being drawn
            {
                if (_geometry.StrokeContains(StrokePen, hitTestParameters.HitPoint, _hitTestTolerance, ToleranceType.Absolute)) //try another hit test with larger tolerance
                    result = new PointHitTestResult(this, hitTestParameters.HitPoint); //if it passed, set it
            }

            return result;
        }

        protected override Geometry DefiningGeometry
        {
            get
            {
                var geometry = new GeometryGroup();
                var collection = new GeometryCollection {DrawLine()};

                if (!string.IsNullOrEmpty(SourceText)) //render text only when its a non-empty string
                    collection.Add(DrawSourceText());
                if (!string.IsNullOrEmpty(MiddleText))
                    collection.Add(DrawMiddleText());
                if (!string.IsNullOrEmpty(TargetText))
                    collection.Add(DrawTargetText());

                geometry.Children = collection;

                geometry.Freeze();
                _geometry = geometry;
                return geometry;
            }
        }


        #region Line drawing

        protected virtual Geometry DrawLine()
        {
            var lineGeometry = new StreamGeometry {FillRule = FillRule.Nonzero};
            using (var context = lineGeometry.Open())
            {
                RelativePosition? linePosition = GetLineToTargetPosition();
                var startPoint = new Point(X1, Y1);
                EndPoint = new Point(X2, Y2);

                //if relationship is not connected, adjust its endpoint to fit end shape, if it is do nothing
                SetEndPoint(linePosition);

                context.BeginFigure(startPoint, false, false);

                DrawAdditionalPoints(context);

                //finish the line
                context.LineTo(EndPoint, true, true);

                if (!(DataContext is RelationshipViewModelBase))
                {
                    //no datacontext, this is just for shape editing
                    DrawEndShape(context, EndPoint, RelativePosition.OnTheLeft);
                }
                else if (!(DataContext is TemporaryRelationship) &&
                         !((RelationshipViewModelBase) DataContext).IsConnected)
                {
                    //add appropriate end shape based on concrete type of the line
                    DrawEndShape(context, EndPoint, linePosition);
                }

                context.Close();
            }

            lineGeometry.Freeze();
            return lineGeometry;
        }

        private void SetEndPoint(RelativePosition? linePosition)
        {
            var baseRelationship = DataContext as RelationshipViewModelBase;
            if (baseRelationship != null && !baseRelationship.IsConnected)
            {
                switch (linePosition)
                {
                    case RelativePosition.OnTheLeft:
                        EndPoint = new Point(X2 - EndShapeSize, Y2);
                        break;
                    case RelativePosition.OnTheRight:
                        EndPoint = new Point(X2 + EndShapeSize, Y2);
                        break;
                    case RelativePosition.Above:
                        EndPoint = new Point(X2, Y2 + EndShapeSize);
                        break;
                    case RelativePosition.Below:
                        EndPoint = new Point(X2, Y2 - EndShapeSize);
                        break;
                    case RelativePosition.Away: //should not occur, do nothing
                    case null:
                        break;
                    default:
                        throw new ArgumentOutOfRangeException();
                }
            }
        }

        private void DrawAdditionalPoints(StreamGeometryContext context)
        {
            //draw additional points (curviness of the line)
            if (AdditionalPoints != null)
            {
                foreach (var point in AdditionalPoints)
                {
                    context.LineTo(point, true, true);
                }
            }
        }

        protected abstract void DrawEndShape(StreamGeometryContext context, Point endShapePoint,
            RelativePosition? linePosition = null);

        #endregion

        #region Text drawing

        private Geometry DrawSourceText()
        {
            var text = new FormattedText(SourceText, CultureInfo.InvariantCulture, FlowDirection.LeftToRight, 
                new Typeface("Calibry"), FontSize, Brushes.Black);
            
                
            var position = GetLineToTargetPosition();
            var textPoint = new Point(X1,Y1);

            if (position == RelativePosition.OnTheLeft || position == null)
            {
                textPoint.Y += 5;
                textPoint.X += 5;
            }
            if (position == RelativePosition.OnTheRight)
            {
                textPoint.Y += 5;
                textPoint.X -= 25;
            }
            if (position == RelativePosition.Above)
            {
                textPoint.Y -= 15;
                textPoint.X += 5;
            }
            if (position == RelativePosition.Below)
            {
                textPoint.Y += 5;
                textPoint.X += 5;
            }

            var geometry = text.BuildGeometry(textPoint);

            geometry.Freeze();
            return geometry;
        }

        private Geometry DrawTargetText()
        {
            var text = new FormattedText(TargetText, CultureInfo.InvariantCulture, FlowDirection.LeftToRight,
                new Typeface("Calibry"), FontSize, Brushes.Black);
            var position = GetLineToTargetPosition();
            var textPoint = EndPoint;

            if (position == RelativePosition.OnTheLeft || position == null)
            {
                textPoint.Y += 5;
                textPoint.X -= 25;
            }
            if (position == RelativePosition.OnTheRight)
            {
                textPoint.Y += 5;
                textPoint.X += 5;
            }
            if (position == RelativePosition.Above)
            {
                textPoint.Y += 5;
                textPoint.X += 5;
            }
            if (position == RelativePosition.Below)
            {
                textPoint.Y -= 20;
                textPoint.X += 5;
            }

            var geometry = text.BuildGeometry(textPoint);

            geometry.Freeze();
            return geometry;
        }

        private Geometry DrawMiddleText()
        {
            var text = new FormattedText(MiddleText, CultureInfo.InvariantCulture, FlowDirection.LeftToRight,
                new Typeface("Calibry"), FontSize, Brushes.Black);

            Geometry geometry;
            if (AdditionalPoints != null && AdditionalPoints.Any())
                //if the path contains any right angle, don't rotate the text
                geometry = text.BuildGeometry(new Point(AdditionalPoints.First().X + 5, AdditionalPoints.First().Y));
            else
            {
                geometry = text.BuildGeometry(new Point((X1 + X2)/2, (Y1 + Y2)/2));
            }

            geometry.Freeze();
            return geometry;
        }

        private Tuple<FlowDirection, double> CalculateTextsDirectionAndTransform(double angle,
            RelationshipTextType textType)
        {
            if (angle >= 90 && angle <= 180)
                if (textType != RelationshipTextType.End)
                    return new Tuple<FlowDirection, double>(FlowDirection.LeftToRight, angle - 180);
                else
                    return new Tuple<FlowDirection, double>(FlowDirection.RightToLeft, angle - 180);
            if (angle >= 0 && angle < 90)
                if (textType != RelationshipTextType.End)
                    return new Tuple<FlowDirection, double>(FlowDirection.RightToLeft, angle);
                else
                    return new Tuple<FlowDirection, double>(FlowDirection.LeftToRight, angle);
            if (angle < 0 && angle >= -90)
                if (textType != RelationshipTextType.End)
                    return new Tuple<FlowDirection, double>(FlowDirection.RightToLeft, angle);
                else
                    return new Tuple<FlowDirection, double>(FlowDirection.LeftToRight, angle);
            if (angle < -90 && angle >= -180)
                if (textType != RelationshipTextType.End)
                    return new Tuple<FlowDirection, double>(FlowDirection.LeftToRight, angle + 180);
                else
                    return new Tuple<FlowDirection, double>(FlowDirection.RightToLeft, angle + 180);
            return null;
        }

        #endregion

        #region Helper methods

        protected RelativePosition? GetLineToTargetPosition()
        {
            var relationship = DataContext as RelationshipViewModelBase;

            //we don't need to draw end shape to temporary relationship (it doesn't have Target property set yet)
            if (relationship == null || relationship is TemporaryRelationship) return null;

            var tempRectangle = new Rectangle();

            //straight line, no angle so we take its start
            if (this.AdditionalPoints == null || !AdditionalPoints.Any())
                tempRectangle.Location = new System.Drawing.Point(X1,Y1);
            else //there is one or more angles, so we take the last one, because it's the one, that matters
                tempRectangle.Location = new System.Drawing.Point(
                    (int) AdditionalPoints.Last().X,
                    (int)AdditionalPoints.Last().Y);

            //set rectangle size plus tolerance, so we can determine it's position to target
            tempRectangle.Width = 1;
            tempRectangle.Height = 1;
            tempRectangle.Inflate(Settings.Default.PathFindingSensitivity, Settings.Default.PathFindingSensitivity);

            //calculate position to end point of the line (it can be connected to another line)
            return tempRectangle.CalculatePositionTo(new Rectangle(X2,Y2,1,1));
        }

        #endregion

        protected virtual void OnMouseLeave(object sender, MouseEventArgs mouseEventArgs)
        {
            var vm = DataContext as RelationshipViewModelBase;
            if (vm == null) return;
            vm.IsHovered = false;

            this.Stroke = Brushes.Black;
            StrokeThickness--;
        }
        protected virtual void OnMouseEnter(object sender, MouseEventArgs mouseEventArgs)
        {
            var vm = DataContext as RelationshipViewModelBase;
            if (vm == null) return;
            vm.IsHovered = true;

            if (!(vm is TemporaryRelationship))
            {
                this.Stroke = Brushes.Red;
            }
            this.StrokeThickness++;
        }

        protected override void OnMouseMove(MouseEventArgs e)
        {
            var vm = DataContext as RelationshipViewModelBase;
            if (vm == null) return;

            //dragging is happening
            if (e.LeftButton == MouseButtonState.Pressed)
            {
                vm.IsHovered = true;
                var mousePosition = e.GetPosition(DiagramCanvas.Instance.DiagramItemsControl);
                var moveTarget = IsPointCloseToAnyOtherPoint(mousePosition, new[] { vm.TargetPoint });
                var moveSource = IsPointCloseToAnyOtherPoint(mousePosition, new[] {vm.SourcePoint});
                if (moveTarget.HasValue && moveSource.HasValue)
                {
                    double sourceDistance = mousePosition.CalculateDistance(moveSource.Value);
                    double targetDistance = mousePosition.CalculateDistance(moveTarget.Value);
                    if (sourceDistance < targetDistance)
                    {
                        moveTarget = null;
                    }
                    else
                    {
                        moveSource = null;
                    }
                }
                if (moveTarget.HasValue) //check if target point should be moved
                {
                    MoveTarget(mousePosition,vm);
                } else if (moveSource.HasValue)
                {
                    MoveSource(mousePosition, vm);
                }
            }
        }

        private void MoveTarget(Point toPosition, RelationshipViewModelBase relationship)
        {
            var position = GetLineToTargetPosition();
            Point adjustedPoint;
            switch (position)
            {
                case RelativePosition.OnTheLeft:
                case RelativePosition.OnTheRight:
                    double top = toPosition.Y - Y2;
                    adjustedPoint = new Point(relationship.TargetPoint.X, relationship.TargetPoint.Y + top);
                    if (CanMoveVertically(adjustedPoint, relationship.Target.Rectangle))
                    {
                        if (relationship.AdditionalPoints.Any())
                        {
                            relationship.TargetPoint = adjustedPoint;
                            var tempAdditionalPoint = relationship.AdditionalPoints[relationship.AdditionalPoints.Count - 1];
                            relationship.AdditionalPoints[relationship.AdditionalPoints.Count - 1] =
                                new Point(tempAdditionalPoint.X, tempAdditionalPoint.Y + top);
                        }
                        else if(CanMoveVertically(adjustedPoint,relationship.Source.Rectangle))
                        {
                            relationship.TargetPoint = adjustedPoint;
                            var sourceAdjusted = new Point(relationship.SourcePoint.X, relationship.TargetPoint.Y);
                            relationship.SourcePoint = sourceAdjusted;
                        }
                    }
                    break;
                case RelativePosition.Above:
                case RelativePosition.Below:
                    double left = toPosition.X - X2;
                    adjustedPoint = new Point(relationship.TargetPoint.X + left, relationship.TargetPoint.Y);
                    if (CanMoveHorizontally(adjustedPoint, relationship.Target.Rectangle))
                    {
                        if (relationship.AdditionalPoints.Any())
                        {
                            relationship.TargetPoint = adjustedPoint;
                            var tempAdditionalPoint = relationship.AdditionalPoints [relationship.AdditionalPoints.Count - 1];
                            relationship.AdditionalPoints[relationship.AdditionalPoints.Count - 1] =
                                new Point(tempAdditionalPoint.X + left, tempAdditionalPoint.Y);
                        }
                        else if(CanMoveHorizontally(adjustedPoint,relationship.Source.Rectangle))
                        {
                            relationship.TargetPoint = adjustedPoint;
                            var sourceAdjusted = new Point(relationship.TargetPoint.X, relationship.SourcePoint.Y);
                            relationship.SourcePoint = sourceAdjusted;
                        }
                    }
                    break;
                case RelativePosition.Away:
                case null:
                    break;
                default:
                    throw new ArgumentOutOfRangeException();
            }
        }

        private void MoveSource(Point toPosition, RelationshipViewModelBase relationship)
        {
            var position = GetLineToTargetPosition();
            Point adjustedPoint;
            switch (position)
            {
                case RelativePosition.OnTheLeft:
                case RelativePosition.OnTheRight:
                    double top = toPosition.Y - Y1;
                    adjustedPoint = new Point(relationship.SourcePoint.X, toPosition.Y);
                    if (CanMoveVertically(adjustedPoint, relationship.Source.Rectangle))
                    {
                        if (relationship.AdditionalPoints.Any())
                        {
                            relationship.SourcePoint = adjustedPoint;
                            var tempAdditionalPoint = relationship.AdditionalPoints[0];
                            relationship.AdditionalPoints[0] = new Point(tempAdditionalPoint.X, tempAdditionalPoint.Y + top);
                        }
                        else if(CanMoveVertically(adjustedPoint,relationship.Target.Rectangle))
                        {
                            relationship.SourcePoint = adjustedPoint;
                            var targetAdjusted = new Point(relationship.TargetPoint.X, relationship.SourcePoint.Y);
                            relationship.TargetPoint = targetAdjusted;
                        }
                    }
                    break;
                case RelativePosition.Above:
                case RelativePosition.Below:
                    double left = toPosition.X - X1;
                    adjustedPoint = new Point(toPosition.X, relationship.SourcePoint.Y);
                    if (CanMoveHorizontally(adjustedPoint, relationship.Source.Rectangle))
                    {
                        if(relationship.AdditionalPoints.Any())
                        {
                            relationship.SourcePoint = adjustedPoint;
                            var tempAdditionalPoint = relationship.AdditionalPoints[0];
                            relationship.AdditionalPoints[0] = new Point(tempAdditionalPoint.X + left,
                                tempAdditionalPoint.Y);
                        }
                        else if(CanMoveHorizontally(adjustedPoint,relationship.Target.Rectangle))
                        {
                            relationship.SourcePoint = adjustedPoint;
                            var targetAdjusted = new Point(relationship.SourcePoint.X, relationship.TargetPoint.Y);
                            relationship.TargetPoint = targetAdjusted;
                        }
                    }
                    break;
                case RelativePosition.Away:
                case null:
                    break;
                default:
                    throw new ArgumentOutOfRangeException();
            }
        }

        private Point? IsPointCloseToAnyOtherPoint(Point thePoint, ICollection<Point> closeTo)
        {
            var closestPoint = closeTo.GetClosestPoint(thePoint);
            if (closestPoint.CalculateDistance(thePoint) <= Math.Pow(_hitTestTolerance,_hitTestTolerance))
            {
                return closestPoint;
            }
            return null;
        }

        private bool CanMoveHorizontally(Point point, Rectangle rectangle)
        {
            return (point.X >= rectangle.Left) && (point.X <= (rectangle.Left + rectangle.Width));
        }

        private bool CanMoveVertically(Point point, Rectangle rectangle)
        {
            return (point.Y >= rectangle.Top) && (point.Y <= (rectangle.Top + rectangle.Height));
        }
    }
}
