﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Windows.Input;
using UmlDiagramDesigner.Core;
using UmlDiagramDesigner.Models.DiagramValidation;
using UmlDiagramDesigner.Services;
using UmlDiagramDesigner.ViewModels.Commands;
using UmlDiagramDesigner.ViewModels.DependentElements;
using UmlDiagramDesigner.ViewModels.DiagramItems;
using UmlDiagramDesigner.ViewModels.DiagramRelationships;

namespace UmlDiagramDesigner.ViewModels.InputForms
{
    public class ClassEditViewModel : InputFormViewModelBase
    {
        private string _name;
        private AttributeViewModel _selectedAttribute;
        private ObservableCollection<AttributeViewModel> _attributes;
        private ObservableCollection<TriggerViewModel> _triggers;
        private TriggerViewModel _selectedTrigger;
        private ClassDiagramItemViewModel _class;
        
        private readonly EntityValidator _validator = new EntityValidator();
        private readonly IDialogService _dialogService;

        public ClassEditViewModel(ClassDiagramItemViewModel @class)
        {
            _class = @class;
            _dialogService = new WpfDialogService();
			RemovedAttributes = new List<AttributeViewModel>();
        }

        #region Properties

        public string Name
        {
            get { return _name; }
            set
            {
                if (value == _name) return;
                _name = value;
                OnPropertyChanged();
            }
        }

        public ObservableCollection<AttributeViewModel> Attributes
        {
            get { return _attributes; }
           set { _attributes = value; OnPropertyChanged();}
        }

        public ObservableCollection<TriggerViewModel> Triggers
        {
            get { return _triggers; }
            set { _triggers = value;
                OnPropertyChanged();
            }
        }

        public AttributeViewModel SelectedAttribute
        {
            get { return _selectedAttribute; }
            set { _selectedAttribute = value; OnPropertyChanged();}
        }

        public TriggerViewModel SelectedTrigger
        {
            get { return _selectedTrigger; }
            set { _selectedTrigger = value; OnPropertyChanged();}
        }

	    public IList<AttributeViewModel> RemovedAttributes { get; private set; }

        #endregion

        #region Commands

        public ICommand AddAttributeCommand { get { return new GenericCommand(()=> Attributes.Add(new AttributeViewModel
        {
            Name = string.Format("AttributeName{0}",Attributes.Count + 1)
        }));} }

        public ICommand AddTriggerCommand { get {return new GenericCommand(() => Triggers.Add(new TriggerViewModel
        {
            Name = string.Format("triggerName{0}", Triggers.Count + 1)
        }));}}

        public GenericCommand<AttributeViewModel> SetPrimaryKeyConstraintCommand
        {
            get
            {
                return new GenericCommand<AttributeViewModel>((att) =>
                {
                    if (att == null) return;

                    if (_class.GetIncomingRelationships().Any(rel => rel is IForeignKeyRelationship))
                    {
                        _dialogService.Display(
                            new InvalidOperationException(
                                "Cannot change primary key constraint as long as there are any relationships, which rely on it. Remove them first"));
						//a little hack, so the checkbox will bind back to the proper data value
						att.IsPrimaryKey = att.IsPrimaryKey;
					}
                    else
                    {
						var currentPrimaryKey = Attributes.FirstOrDefault(attribute => attribute.IsPrimaryKey);
						if (!att.IsPrimaryKey && currentPrimaryKey != null) //user tries to set multiple primary keys, set other key IsPrimary to false
						{
							currentPrimaryKey.IsPrimaryKey = false;
							att.IsPrimaryKey = true;
						}
						else
						{
							//a little hack, so the checkbox will bind back to the proper data value
							att.IsPrimaryKey = !att.IsPrimaryKey;
						}
					}
                });
            }
        }

        public GenericCommand<AttributeViewModel> DeleteAttributeCommand
        {
            get { return new GenericCommand<AttributeViewModel>(att =>
            {
                if (att == null) return;

                //if user wants to remove primary key, check for IForeignKeyRelationships, because they have to be removed first
                if (att.IsPrimaryKey && _class.GetIncomingRelationships().Any(rel => rel is IForeignKeyRelationship))
                {
                    _dialogService.Display(
                        new ArgumentException(
                            "Cannot remove primary key, because there are relationships which depend on it. Remove them first."));
                    return;
                }
                Attributes.Remove(att);
				RemovedAttributes.Add(att);
            });}
        }

        public GenericCommand<TriggerViewModel> DeleteTriggerCommand
        {
            get { return new GenericCommand<TriggerViewModel>(trigger =>
            {
                if (trigger == null) return;
                Triggers.Remove(trigger);
            });}
        }

        #endregion

        protected override bool ValidateInputData()
        {
            //get hierarchy in both directions (without duplicities - Union)
            var hierarchy = _class.DepthFirstTraversal<InheritanceViewModel>()
                                   .Union(_class.DepthFirstTraversal<InheritanceViewModel>(true))
                                   .ToArray();

            ValidationResult result;

            if (!(result = _validator.ValidateName(Name)).Successful)
            {
                return SetErrorText(result.ErrorMessage);
            }

            if (!(result = _validator.Validate(hierarchy.OfType<ClassDiagramItemViewModel>()
                                                        .SelectMany(entity => entity.Attributes)
                                                        .Union(Attributes)
                                                        .ToArray())).Successful)
            {
                return SetErrorText(result.ErrorMessage);
            }

            if (!(result = _validator.Validate(hierarchy.OfType<ClassDiagramItemViewModel>()
                                                        .SelectMany(entity => entity.Triggers)
                                                        .Union(Triggers)
                                                        .ToArray())).Successful)
            {
                return SetErrorText(result.ErrorMessage);
            }

            return true;
        }
    }
}
