﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Windows.Media;
using UmlDiagramDesigner.Core;
using UmlDiagramDesigner.Core.Enums;
using UmlDiagramDesigner.Helpers;
using UmlDiagramDesigner.Models.DiagramValidation;
using UmlDiagramDesigner.Models.Shapes;
using UmlDiagramDesigner.Services;
using UmlDiagramDesigner.ViewModels.DependentElements;
using UmlDiagramDesigner.ViewModels.DiagramItems;
using UmlDiagramDesigner.ViewModels.DiagramRelationships;

namespace UmlDiagramDesigner.ViewModels.InputForms
{
    public class RelationshipInputViewModel : InputFormViewModelBase
    {
        private string _sourceName;
        private string _targetName;
        private MultiplicityType _targetMultiplicity;
        private MultiplicityType _sourceMultiplicity;
        private string _middleText;
        private readonly Type _relationshipType;

        private readonly Dictionary<Type, Func<RelationshipShapeBase>> _shapesByType;
        private AttributeViewModel _targetPrimaryKey;
        private AttributeViewModel _sourceForeignKey;
        private readonly EntityValidator _entityValidator = new EntityValidator();
        private readonly RelationshipViewModelBase _relationship;
        private readonly IDialogService _dialogService;

        public RelationshipInputViewModel()
        {
            
        }

        public RelationshipInputViewModel(RelationshipViewModelBase relationship)
        {
            _relationship = relationship;
            _dialogService = new WpfDialogService();

            if (relationship.Target != null)
            {
                TargetName = relationship.Target.Name;

                var entity = relationship.Target as ClassDiagramItemViewModel;
                if (entity != null)
                {
                    InitializeKeyConstraints(relationship, entity);
                }
            }
            
            _shapesByType = CreateShapesByTypeDictionary();
            SourceName = relationship.Source.Name;
            SourceMultiplicity = relationship.SourceMultiplicity;
            TargetMultiplicity = relationship.TargetMultiplicity;
            MiddleText = relationship.MiddleText ?? string.Empty;
            _relationshipType = relationship.GetType();
        }

        private void InitializeKeyConstraints(RelationshipViewModelBase relationship, ClassDiagramItemViewModel @class)
        {
            TargetPrimaryKey = @class.Attributes.FirstOrDefault(att => att.IsPrimaryKey);
            if (((IForeignKeyRelationship) relationship).SourceForeignKey != null)
            {
                //there is already a foreign key (we are just editing something)
                SourceForeignKey = ((IForeignKeyRelationship) relationship).SourceForeignKey;
            }
            else if (TargetPrimaryKey != null)
            {
                //there is no foreign key, create a default one
                SourceForeignKey = new AttributeViewModel
                {
                    Name = string.Format("{0}{1}", @class.Name, TargetPrimaryKey.Name),
                    DataType = TargetPrimaryKey.DataType,
                    IsForeignKey = true
                };
            }
            else
            {
                SourceForeignKey = new AttributeViewModel()
                {
                    Name = string.Format("{0}Id", @class.Name),
                    DataType = DbDataType.Integer,
                    IsForeignKey = true
                };
            }
        }

        #region Properties

        public string SourceName
        {
            get { return _sourceName; }
            set { _sourceName = value; OnPropertyChanged();}
        }

        public string TargetName
        {
            get { return _targetName; }
            set { _targetName = value; OnPropertyChanged();}
        }

        public AttributeViewModel SourceForeignKey
        {
            get { return _sourceForeignKey; }
            set { _sourceForeignKey = value; OnPropertyChanged();}
        }

        public AttributeViewModel TargetPrimaryKey
        {
            get { return _targetPrimaryKey; }
            set { _targetPrimaryKey = value; OnPropertyChanged();}
        }

        public RelationshipShapeBase Relationship
        {
            get { return CreateRelationshipShape(); }
        }

        public MultiplicityType SourceMultiplicity
        {
            get { return _sourceMultiplicity; }
            set { _sourceMultiplicity = value; OnPropertyChanged(); OnPropertyChanged("Relationship"); }
        }

        public MultiplicityType TargetMultiplicity
        {
            get { return _targetMultiplicity; }
            set { _targetMultiplicity = value; OnPropertyChanged(); OnPropertyChanged("Relationship");}
        }

        public string MiddleText
        {
            get { return _middleText; }
            set { _middleText = value; OnPropertyChanged(); OnPropertyChanged("Relationship"); }
        }

        #endregion

        private Dictionary<Type, Func<RelationshipShapeBase>> CreateShapesByTypeDictionary()
        {
            var dict = new Dictionary<Type, Func<RelationshipShapeBase>>()
            {
                {typeof(AssociationViewModel),() => new AssociationShape()},
                {typeof(AggregationViewModel),() => new AggregationShape()},
                {typeof(CompositionViewModel),() => new CompositionShape()}
            };

            return dict;
        }

        private RelationshipShapeBase CreateRelationshipShape()
        {
            if (!_shapesByType.ContainsKey(_relationshipType)) return null;

            var shape = _shapesByType[_relationshipType]();

            shape.X1 = 20;
            shape.Y1 = 35;
            shape.X2 = 440;
            shape.Y2 = 35;

            shape.Stroke = Brushes.Black;
            shape.StrokeThickness = 1;

            shape.SourceText = SourceMultiplicity.GetString();
            shape.MiddleText = MiddleText;
            shape.TargetText = TargetMultiplicity.GetString();

            shape.FontSize = 13;

            return shape;
        }

        protected override bool ValidateInputData()
        {
            var source = _relationship.Source as ClassDiagramItemViewModel;

            if (source != null)
            {
                ValidationResult result;

                //union all attributes from source except the SourceForeignKey (because it would throw a duplicate name error)
                if (!(result = _entityValidator.Validate(source.Attributes.Where(att => att != SourceForeignKey)
                                                                          .Union(new[] {SourceForeignKey})
                                                                          .ToArray())).Successful)
                {
                    _dialogService.Display(result.ErrorMessage);
                    return false;
                }
            }

            return true;
        }

        protected override bool CanCancel()
        {
            if (_relationship is IForeignKeyRelationship)
            {
                //if relationship defines a foreign key, it has to be set and valid!
                return SourceForeignKey != null &&
                       SourceForeignKey.IsValid() &&
                       ((ClassDiagramItemViewModel)_relationship.Source).Attributes.Contains(SourceForeignKey);
            }
            //otherwise we can cancel without any problem
            return true;
        }
    }
}
