﻿using System;
using System.Collections.ObjectModel;
using System.Windows;
using System.Windows.Input;
using UmlDiagramDesigner.Core;
using UmlDiagramDesigner.DAL.Database;
using UmlDiagramDesigner.DTO;
using UmlDiagramDesigner.Helpers;
using UmlDiagramDesigner.Services;
using UmlDiagramDesigner.ViewModels.Commands;

namespace UmlDiagramDesigner.ViewModels
{
    public class SavedDiagramsViewModel : ViewModelBase
    {
        private string _diagramName;
        private CanvasViewModel _diagramToSave;
        private bool _isDbInitialized;
        private DiagramNameAndItemCountDTO _selectedDiagram;
        private ObservableCollection<DiagramNameAndItemCountDTO> _savedDiagrams;

        private readonly DiagramPersistanceFacade _diagramPersistance = new DiagramPersistanceFacade();
        private readonly IDialogService _dialog = new WpfDialogService();

        #region Constructors

        /// <summary>
        /// Initializes a new instance of the <see cref="SavedDiagramsViewModel"/> class.
        /// Use this constructor for loading diagrams, not saving.
        /// </summary>
        public SavedDiagramsViewModel()
        {
            if(!DatabaseHelper.Instance.IsInitialized)
                IsDbInitialized = false;
            else
            {
                IsDbInitialized = true;
                FetchDiagramsFromDatabase();
            }

            DatabaseHelper.Instance.InitializationCompleted += (sender, args) =>
            {
                IsDbInitialized = true;
                FetchDiagramsFromDatabase();
            };
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="SavedDiagramsViewModel"/> class.
        /// Use this overload for saving diagram.
        /// </summary>
        /// <param name="diagram">The diagram.</param>
        public SavedDiagramsViewModel(CanvasViewModel diagram)
        {
            if (!DatabaseHelper.Instance.IsInitialized)
                IsDbInitialized = false;
            else
            {
                IsDbInitialized = true;
                FetchDiagramsFromDatabase();
            }

            DatabaseHelper.Instance.InitializationCompleted += (sender, args) =>
            {
                IsDbInitialized = true;
                FetchDiagramsFromDatabase();
            };

            _diagramToSave = diagram;
        }

        #endregion

        #region Properties

        public Window SavedDiagramsWindow { get; set; }

        public bool IsDbInitialized
        {
            get { return _isDbInitialized; }
            set { _isDbInitialized = value; OnPropertyChanged();}
        }

        public string DiagramName
        {
            get { return _diagramName; }
            set
            {
                _diagramName = value;
                OnPropertyChanged();
            }
        }

        public ObservableCollection<DiagramNameAndItemCountDTO> SavedDiagrams
        {
            get { return _savedDiagrams; }
            set { _savedDiagrams = value; OnPropertyChanged();}
        }

        public DiagramNameAndItemCountDTO SelectedDiagram
        {
            get { return _selectedDiagram; }
            set { _selectedDiagram = value;
                OnPropertyChanged();
                OnPropertyChanged("IsDeleteEnabled");
                OnPropertyChanged("IsLoadEnabled");

                if (SelectedDiagram != null && !string.IsNullOrEmpty(SelectedDiagram.Name) && IsSaveEnabled)
                {
                    DiagramName = SelectedDiagram.Name;
                }
            }
        }

        public bool IsSaveEnabled
        {
            get { return _diagramToSave != null; }
        }

        public bool IsDeleteEnabled { get { return SelectedDiagram != null; } }

        public bool IsLoadEnabled { get { return SelectedDiagram != null; } }

        #endregion

        #region Commands

        public ICommand SaveDiagramCommand { get {return new GenericCommand(SaveDiagram);} }

        public ICommand LoadDiagramCommand { get {return new GenericCommand(LoadDiagram);}}

        public ICommand DeleteDiagramCommand { get {return new GenericCommand(DeleteDiagram);}}

        public ICommand RefreshDataCommand { get { return new GenericCommand(RefreshData);}}

        #endregion

        #region Private methods

        private void SaveDiagram()
        {
            if (_diagramToSave == null || string.IsNullOrEmpty(_diagramName))
            {
                _dialog.Display("Set the diagram name you want to save.");
                return;
            }

            _diagramToSave.Name = DiagramName;

            if (_diagramPersistance.Save(_diagramToSave))
            {
                _diagramToSave = null;
                OnPropertyChanged("IsSaveEnabled");
                DiagramName = string.Empty;
                FetchDiagramsFromDatabase();
            }
        }

        private void LoadDiagram()
        {
            if (SelectedDiagram == null) return;

            var diagram = _diagramPersistance.Load(SelectedDiagram.Id);

            if (diagram == null) return;

            RaiseDiagramLoadedEvent(diagram);

            if(SavedDiagramsWindow != null)
                SavedDiagramsWindow.Close();
        }

        private void DeleteDiagram()
        {
            if (SelectedDiagram == null) return;

            if (_diagramPersistance.Delete(SelectedDiagram.Id))
            {
                SelectedDiagram = null;
                FetchDiagramsFromDatabase();
            }
        }

        private void FetchDiagramsFromDatabase()
        {
            var diagrams = _diagramPersistance.GetAllAvailable();

            if (diagrams == null) return;

            SavedDiagrams = new ObservableCollection<DiagramNameAndItemCountDTO>(diagrams);
        }

        private void RaiseDiagramLoadedEvent(CanvasViewModel diagram)
        {
            if(DiagramLoaded != null)
                DiagramLoaded(this, diagram);
        }

        private void RaiseDiagramSavedEvent(CanvasViewModel diagram)
        {
            if(DiagramSaved != null)
                DiagramSaved(this,diagram);
        }

        private void RefreshData()
        {
            if (!DatabaseHelper.Instance.IsInitialized)
                DatabaseHelper.Instance.InitializeDbAsync();
            else
                FetchDiagramsFromDatabase();
        }

        #endregion


        public event EventHandler<CanvasViewModel> DiagramLoaded;
        public event EventHandler<CanvasViewModel> DiagramSaved;
    }
}
