﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Text;
using System.Threading.Tasks;
using UmlDiagramDesigner.CodeGeneration.Contracts;
using UmlDiagramDesigner.CodeGeneration.Enums;
using UmlDiagramDesigner.Core.Enums;
using UmlDiagramDesigner.Data;

namespace UmlDiagramDesigner.CodeGeneration.ConcreteGenerators
{
	/// <summary>
	/// Concrete C# generator.
	/// </summary>
	public class CSharpGenerator : ObjectOrientedLanguageGenerator
    {
        private IList<string> _requiredNamespaces;

		/// <summary>
		/// Initializes a new instance of the <see cref="CSharpGenerator"/> class.
		/// </summary>
		/// <param name="languageProperties">The language properties.</param>
		public CSharpGenerator(IOopLanguageProperties languageProperties)
            : base(languageProperties)
        {

        }

        protected override IList<string> RequiredNamespaces
        {
            get { return _requiredNamespaces ?? (_requiredNamespaces = new []
            {
                "System",
                "System.Collections.Generic",
            }); }
        }
 
        public override string LanguageName
        {
            get { return "C#"; }
        }

        public override string FileExtension
        {
            get { return ".cs"; }
        }

        protected override string GenerateMainContent(Diagram diagram)
        {
            var sb = new StringBuilder();

            sb.AppendLine(GenerateNamespaceImports());

            sb.AppendLine(GenerateNamespaceStart(diagram.Name));

            foreach (var @class in diagram.Classes)
            {
                sb.AppendLine(GenerateClass(@class).InsertTabsAtStartOfEachLine());
            }

            sb.AppendLine(GenerateNamespaceEnd());

            return sb.ToString();
        }
        
        protected override string GenerateClass(Class @class)
        {
            var sb = new StringBuilder();

            sb.AppendLine(GenerateClassName(@class));
            sb.AppendLine(GenerateStartBlock());

            sb.AppendLine(GenerateConstructor(@class).InsertTabsAtStartOfEachLine());

            sb.AppendLine(GenerateAttributes(@class).InsertTabsAtStartOfEachLine());

            sb.AppendLine(GenerateEndBlock());

            return sb.ToString();
        }

		/// <summary>
		/// Generates the constructor.
		/// </summary>
		/// <param name="class">The class.</param>
		/// <returns></returns>
		private string GenerateConstructor(Class @class)
        {
            string publicAccessibility;
            LanguageProperties.Accessibilities.TryGetValue(Accessibility.Public, out publicAccessibility);

            if (@class.Compositions.Count == 0)
            {
                //vygenerovat prázdný konstruktor
                return string.Format("{0} {1}() {2}{3}", publicAccessibility ?? string.Empty, @class.Name,
                    GenerateStartBlock(), GenerateEndBlock());
            }

            var sb = new StringBuilder();
            sb.AppendLine(string.Format("{0} {1}(", publicAccessibility ?? string.Empty, @class.Name));
            foreach (var composition in @class.Compositions)
            {
                //přidat povinné parametry do konstruktoru
                sb.Append(string.Format("{0} {1},", LanguageProperties.GenerateNavPropertyDataType(composition),
                    DetermineNavPropertyName(composition).ToCamelCase()));
            }

            sb.Remove(sb.Length - 1,1);//smazat poslední ','

            sb.Append(string.Format(") \n {0}", GenerateStartBlock()));

            sb.Append(GenerateConstructorInitializations(@class.Compositions).InsertTabsAtStartOfEachLine());

            sb.AppendLine(GenerateEndBlock());

            return sb.ToString();
        }

		/// <summary>
		/// Generates the constructor initializations.
		/// </summary>
		/// <param name="attributes">The attributes.</param>
		/// <returns></returns>
		private string GenerateConstructorInitializations(IEnumerable<ForeignKeyConstraint> attributes)
        {
            var sb = new StringBuilder();

            foreach (var attribute in attributes)
            {
                sb.AppendLine(string.Format("this.{0} = {1}{2}", 
                    DetermineNavPropertyName(attribute).ToCamelCase(),
                    DetermineNavPropertyName(attribute).ToCamelCase(),
                    GenerateLineEnd()));
            }

            return sb.ToString();
        }

		/// <summary>
		/// Generates the attributes.
		/// </summary>
		/// <param name="class">The class.</param>
		/// <returns></returns>
		private string GenerateAttributes(Class @class)
        {
            var sb = new StringBuilder();
            string comment, publicAccessibility;
            LanguageProperties.Features.TryGetValue(LanguageFeature.Comment, out comment);
            LanguageProperties.Accessibilities.TryGetValue(Accessibility.Public, out publicAccessibility);


            if (@class.PrimaryKey != null)
            {
                sb.AppendLine(string.Format("{0} {1} {2} {3}", publicAccessibility ?? string.Empty,
                    LanguageProperties.DataTypes[@class.PrimaryKey.DataType],
                    @class.PrimaryKey.Name,
                    "{ get; set; }"));
            }

            if (@class.Attributes.Any())
            {
                sb.AppendLine();
                sb.AppendLine(string.Format("{0} Generated attributes.", comment));

                foreach (var attribute in @class.Attributes)
                {
                    sb.AppendLine(string.Format("{0} {1} {2} {3}", publicAccessibility ?? string.Empty,
                        LanguageProperties.DataTypes[attribute.DataType],
                        attribute.Name,
                        "{ get; set; }"));
                }
            }

            if (@class.Aggregations.Any() || @class.Associations.Any() || @class.Compositions.Any())
            {
                sb.AppendLine();
                sb.AppendLine(string.Format("{0} Generated navigation properties.", comment));

                foreach (var navigationAttribute in @class.Aggregations.Union(@class.Associations).Union(@class.Compositions))
                {
                    sb.AppendLine(string.Format("{0} {1} {2} {3}", publicAccessibility ?? string.Empty,
                        LanguageProperties.GenerateNavPropertyDataType(navigationAttribute),
                        DetermineNavPropertyName(navigationAttribute),
                        "{ get; set; }"));
                }
            }

            return sb.ToString();
        }
    }
}
