﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using UmlDiagramDesigner.CodeGeneration.Contracts;
using UmlDiagramDesigner.CodeGeneration.Enums;
using UmlDiagramDesigner.Data;

namespace UmlDiagramDesigner.CodeGeneration.ConcreteGenerators
{
	/// <summary>
	/// Concrete generator for Smalltalk language.
	/// </summary>
	public class SmallTalkGenerator : ObjectOrientedLanguageGenerator
    {
		/// <summary>
		/// Initializes a new instance of the <see cref="SmallTalkGenerator"/> class.
		/// </summary>
		/// <param name="languageProperties">The language properties.</param>
		public SmallTalkGenerator(IOopLanguageProperties languageProperties) : base(languageProperties)
        {
        }

        public override string LanguageName
        {
            get { return "SmallTalk"; }
        }

        public override string FileExtension
        {
            get { return ".st"; }
        }

        protected override string GenerateMainContent(Diagram diagram)
        {
            var sb = new StringBuilder();

            sb.AppendLine();

            //no need for namespaces or anything, just generate all classes
            foreach (var @class in diagram.Classes.OrderBy(cl => cl.Ancestor != null))
            {
                sb.AppendLine(GenerateClass(@class));
            }

            return sb.ToString();

        }

        protected override IList<string> RequiredNamespaces
        {
            get { return new List<string>(); }
        }

        protected override string GenerateClass(Class @class)
        {
            var sb = new StringBuilder();

            if (@class.Ancestor == null) //inherit from Object class
            {
                sb.AppendLine(string.Format("Object subclass: #{0}",@class.Name));
                sb.Append(string.Format("instanceVariableNames: ' {0}'",
                    GenerateAttributes(@class)).InsertTabsAtStartOfEachLine());
                sb.Append(string.Format("classVariableNames: ''").InsertTabsAtStartOfEachLine());
                sb.Append(string.Format("poolDictionaries: ''").InsertTabsAtStartOfEachLine());
                sb.Append(string.Format("category: 'nil'!").InsertTabsAtStartOfEachLine());
            }
            else
            {
                sb.AppendLine(string.Format("{0} subclass: #{1}", @class.Ancestor.Name, @class.Name));
                sb.Append(string.Format("instanceVariableNames: ' {0}'",
                    GenerateAttributes(@class)).InsertTabsAtStartOfEachLine());
                sb.Append(string.Format("classVariableNames: ''").InsertTabsAtStartOfEachLine());
                sb.Append(string.Format("poolDictionaries: ''").InsertTabsAtStartOfEachLine());
                sb.Append(string.Format("category: 'nil'!").InsertTabsAtStartOfEachLine());
            }

            return sb.ToString();
        }

		public string GenerateAttributes(Class @class)
        {
            var sb = new StringBuilder();

            if (@class.PrimaryKey != null)
            {
                sb.Append(string.Format("{0} ", @class.PrimaryKey.Name));
            }

            foreach (var attribute in @class.Attributes)
            {
                sb.Append(string.Format("{0} ", attribute.Name));
            }

            foreach (var source in @class.Aggregations.Union(@class.Associations).Union(@class.Aggregations))
            {
                sb.Append(string.Format("{0} ",source.Name));
            }

            return sb.ToString();
        }

		/// <summary>
		/// Generates the opening part of document. It contains only information data about generated code.
		/// Has to use different code because of comment character.
		/// </summary>
		/// <returns></returns>
		protected override string GenerateOpening()
        {
            var sb = new StringBuilder();

            string comment;
            if (LanguageProperties.Features.TryGetValue(LanguageFeature.Comment, out comment))
            {
                sb.AppendLine(string.Format("{0}***************************************************", comment));
                sb.AppendLine();
                sb.AppendLine(
                    string.Format("This code was automatically generated to {0} language from UML diagram on {1}.",
                        LanguageName, DateTime.Now.ToShortDateString()));
                sb.AppendLine(string.Format("Use at your own risk."));
                sb.AppendLine();
                sb.AppendLine(string.Format("***************************************************{0}", comment));
            }
            return sb.ToString();
        }
    }
}