﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using UmlDiagramDesigner.CodeGeneration.Contracts;
using UmlDiagramDesigner.CodeGeneration.Enums;
using UmlDiagramDesigner.Data;

namespace UmlDiagramDesigner.CodeGeneration.ConcreteGenerators
{
	/// <summary>
	/// Generator of SQL language.
	/// </summary>
	public class SqlGenerator : GeneratorBase
    {
		/// <summary>
		/// Initializes a new instance of the <see cref="SqlGenerator"/> class.
		/// </summary>
		/// <param name="tableGenerator">The table generator.</param>
		public SqlGenerator(ISqlTableGenerator tableGenerator) : base(tableGenerator.SqlProperties)
        {
            TableGenerator = tableGenerator;
        }

		/// <summary>
		/// Gets or sets the table generator.
		/// </summary>
		/// <value>
		/// The table generator.
		/// </value>
		public ISqlTableGenerator TableGenerator { get; set; }

        public override string LanguageName
        {
            get { return "SQL"; }
        }

        public override string FileExtension
        {
            get { return ".sql"; }
        }

        protected override string GenerateMainContent(Diagram diagram)
        {
            var sb = new StringBuilder();

            sb.AppendLine(TableGenerator.GenerateTables(diagram.Classes));

            sb.AppendLine();

            sb.AppendLine(GenerateViews(diagram.Views));

            return sb.ToString();
        }

        protected virtual string GenerateViews(IEnumerable<View> views)
        {
            var sb = new StringBuilder();
            bool firstIteration = true;

            foreach (var view in views)
            {
                if (firstIteration)
                {
                    sb.AppendLine();
                    sb.AppendLine(GenerateComment("Generating views"));
                    sb.AppendLine();
                    firstIteration = false;
                }
                sb.AppendLine(GenerateView(view));
            }

            return sb.ToString();
        }

        protected virtual string GenerateView(View view)
        {
            var sb = new StringBuilder();

            sb.AppendLine(string.Format("CREATE VIEW {0} AS", ValidateString(view.Name)));
            sb.AppendLine(string.Format("SELECT {0}", string.Join(",",
                view.Attributes.Select(
                    attribute => string.Format("{0}.{1}", ValidateString(attribute.PartOf.Name), attribute.Name)))));
            sb.AppendLine(string.Format("FROM {0};", string.Join(",", view.Attributes.GroupBy(att => att.PartOf.Name).Select(g => g.Key))));

            return sb.ToString();
        }

		/// <summary>
		/// Generates the comment.
		/// </summary>
		/// <param name="comment">The comment.</param>
		/// <returns></returns>
		protected string GenerateComment(string comment)
        {
            return string.Format("{0} {1}", TableGenerator.SqlProperties.Features[LanguageFeature.Comment], comment);
        }

		/// <summary>
		/// Validates the string. If it contains a keyword, it is wrapped in escape characters.
		/// </summary>
		/// <param name="input">The input.</param>
		/// <returns></returns>
		protected string ValidateString(string input)
        {
            if (TableGenerator.SqlProperties.Keywords.Contains(input.ToLower()))
            {
                return string.Format("{1}{0}{1}", input,
                    TableGenerator.SqlProperties.Features[LanguageFeature.EscapeCharacter]);
            }
            return input;
        }
    }
}
