﻿using System;
using System.Collections.Generic;
using System.Diagnostics.Eventing.Reader;
using System.Linq;
using System.Text;
using UmlDiagramDesigner.CodeGeneration.Contracts;
using UmlDiagramDesigner.CodeGeneration.DTO;
using UmlDiagramDesigner.CodeGeneration.Enums;
using UmlDiagramDesigner.Core;
using UmlDiagramDesigner.Data;

namespace UmlDiagramDesigner.CodeGeneration.ConcreteGenerators
{
    /// <summary>
    /// This strategy will generate the whole inheritance hierarchy into a single database table. 
    /// It's best for performance and simplicity, but may grow chaotic with large hierarchy.
    /// </summary>
    public class TablePerHierarchyGenerator : TableGeneratorBase
    {
        public TablePerHierarchyGenerator(ISqlProperties sqlProperties)
        {
            SqlProperties = sqlProperties;
        }

        public override string GenerateTables(IEnumerable<Class> classes)
        {
            var combinedClasses = classes.Where(@class => @class.Ancestor == null)
                                     .Select(CreateCombinedClass)
                                     .ToArray();

            var sb = new StringBuilder();

            foreach (var rootClass in combinedClasses)
            {
                sb.AppendLine(GenerateSingleTable(rootClass));
            }

            sb.AppendLine();

            foreach (var combinedClass in combinedClasses)
            {
                foreach (var foreignKey in combinedClass.ForeignKeys)
                {
                    var targetClass = combinedClasses.FirstOrDefault(dto => dto.CombinedOf.Contains(foreignKey.References.PartOf));
                    if(targetClass == null) continue;

                    //sb.AppendLine(GenerateForeignKeyConstraint(
                    //    combinedClass.Name, 
                    //    foreignKey.Name, 
                    //    targetClass.Name,
                    //    targetClass.PrimaryKey.Name));
                }
            }

            return sb.ToString();
        }

        private string GenerateSingleTable(CombinedClassDTO combinedClass)
        {
            var sb = new StringBuilder();

            sb.AppendLine(string.Format("CREATE TABLE {0}", ValidateString(combinedClass.Name)));
            sb.AppendLine("(");
            sb.Append(GenerateAttributes(combinedClass).InsertTabsAtStartOfEachLine());
            sb.AppendLine(");");

            return sb.ToString();
        }

        private string GenerateAttributes(CombinedClassDTO combinedClass)
        {
            var sb = new StringBuilder();

            if (combinedClass.PrimaryKey != null)
            {
                sb.AppendLine(string.Format("{0} {1} PRIMARY KEY,", combinedClass.PrimaryKey.Name,
                    SqlProperties.DataTypes[combinedClass.PrimaryKey.DataType]));
            }

            foreach (var attribute in combinedClass.Attributes)
            {
                sb.AppendLine(string.Format("{0} {1},", attribute.Name, SqlProperties.DataTypes[attribute.DataType]));
            }

            foreach (var foreignKey in combinedClass.ForeignKeys)
            {
                sb.AppendLine(string.Format("{0} {1},", foreignKey.Name, SqlProperties.DataTypes[foreignKey.DataType]));
            }

            sb.Remove(sb.Length - 3, 3);

            return sb.ToString();
        }

        private CombinedClassDTO CreateCombinedClass(Class rootClass)
        {
            var hierarchy = rootClass.DepthFirstTraversal().OfType<Class>().ToArray();
            var combinedClass = new CombinedClassDTO()
            {
                Name = rootClass.Name,
                CombinedOf = hierarchy
            };

            foreach (var @class in hierarchy)
            {
                if (@class.PrimaryKey != null)
                {
                    combinedClass.PrimaryKey = @class.PrimaryKey;
                }
                foreach (var attribute in @class.Attributes)
                {
                    combinedClass.Attributes.Add(attribute);
                }
                foreach (var foreignKey in @class.Associations.Union(@class.Aggregations).Union(@class.Compositions))
                {
                    combinedClass.ForeignKeys.Add(foreignKey);
                }
            }

            return combinedClass;
        }
    }
}
