﻿using System;
using System.Collections.Generic;
using System.IO.IsolatedStorage;
using System.Linq;
using System.Text;
using UmlDiagramDesigner.CodeGeneration.Contracts;
using UmlDiagramDesigner.CodeGeneration.Enums;
using UmlDiagramDesigner.Core;
using UmlDiagramDesigner.Core.Enums;
using UmlDiagramDesigner.Data;

namespace UmlDiagramDesigner.CodeGeneration.ConcreteGenerators
{
	/// <summary>
	/// Table per type generation strategy.
	/// </summary>
	public class TablePerTypeGenerator : TableGeneratorBase
    {
		/// <summary>
		/// Initializes a new instance of the <see cref="TablePerTypeGenerator"/> class.
		/// </summary>
		/// <param name="sqlProperties">The SQL properties.</param>
		public TablePerTypeGenerator(ISqlProperties sqlProperties)
        {
            SqlProperties = sqlProperties;
        }

		/// <summary>
		/// Generates the tables. One table per each class plus possible junction tables.
		/// </summary>
		/// <param name="classes">The classes.</param>
		/// <returns></returns>
		/// <exception cref="InvalidCastException">An element in the sequence cannot be cast to type <paramref name="Class" />.</exception>
		/// <exception cref="ArgumentNullException"><paramref name="source" /> is null.</exception>
		public override string GenerateTables(IEnumerable<Class> classes)
        {
            var sb = new StringBuilder();
            var classArray = classes.ToArray();

            //get "root" classes
            foreach (var @class in classArray.Where(c => c.Ancestor == null))
            {
                var hierarchy = @class.DepthFirstTraversal().Cast<Class>().ToArray();
                //iterate the whole hierarchy from top to bottom
                foreach (var classInHierarchy in hierarchy)
                {
                    sb.AppendLine(GenerateTable(classInHierarchy));
                }
            }

            sb.AppendLine(GenerateJunctionTables(classArray));
            sb.AppendLine(GenerateForeignKeys(classArray));
            sb.AppendLine(GenerateTriggers(classArray.SelectMany(c => c.Triggers).ToArray()));

            return sb.ToString();
        }

		/// <summary>
		/// Generates the triggers.
		/// </summary>
		/// <param name="triggers">The triggers.</param>
		/// <returns></returns>
		private string GenerateTriggers(Trigger[] triggers)
        {
            var sb = new StringBuilder();

            if (triggers.Any())
            {
                sb.AppendLine();
                sb.AppendLine(GenerateComment("Generating triggers"));
                sb.AppendLine();
            }

            foreach (var trigger in triggers)
            {
                sb.AppendLine(GenerateTrigger(trigger));
            }

            return sb.ToString();
        }

		/// <summary>
		/// Generates the foreign keys. Takes care of creation junction tables if necessary.
		/// </summary>
		/// <param name="classes">The classes.</param>
		/// <returns></returns>
		private string GenerateForeignKeys(Class[] classes)
        {
            var sb = new StringBuilder();

            bool firstIteration = true;

            foreach (var @class in classes)
            {
                foreach (var foreignKey in @class.Compositions
                    .Union(@class.Aggregations)
                    .Union(@class.Associations)
                    .Where(fk => fk.TargetMultiplicity == MultiplicityType.ZeroToOne || 
                        fk.TargetMultiplicity == MultiplicityType.One))
                {
                    if (firstIteration)
                    {
                        sb.AppendLine();
                        sb.AppendLine(GenerateComment("Generating foreign keys"));
                        sb.AppendLine();
                        firstIteration = false;
                    }
                    sb.AppendLine(GenerateForeignKeyConstraint(foreignKey));
                }

                foreach (var inheritedClass in @class.Children)
                {
                    sb.AppendLine(GenerateForeignKeyConstraint(new ForeignKeyConstraint()
                    { //create temporary primary key that will point to child class
                        Name = string.Format("{0}Id",inheritedClass.Name),
                        DataType = inheritedClass.PrimaryKey.DataType,
                        PartOf = @class,
                        References = inheritedClass.PrimaryKey,
                        SourceMultiplicity = MultiplicityType.One,
                        TargetMultiplicity = MultiplicityType.One
                    }));
                }
            }

            return sb.ToString();
        }

		/// <summary>
		/// Generates the junction tables.
		/// </summary>
		/// <param name="classes">The classes.</param>
		/// <returns></returns>
		private string GenerateJunctionTables(Class[] classes)
        {
            var sb = new StringBuilder();

            bool firstIteration = true;

            foreach (var @class in classes)
            {
                //iterate through all foreign keys, which have both multiplicities of OneToMany or ZeroToMany
                foreach (
                    var foreignKey in @class.Compositions
                    .Union(@class.Aggregations)
                    .Union(@class.Associations)
                    .Where(fk => (fk.SourceMultiplicity == MultiplicityType.OneToMany ||
                                 fk.SourceMultiplicity == MultiplicityType.ZeroToMany) &&
                                 (fk.TargetMultiplicity == MultiplicityType.OneToMany ||
                                 fk.TargetMultiplicity == MultiplicityType.ZeroToMany))
                            )
                {
                    if (firstIteration)
                    {
                        sb.AppendLine();
                        sb.AppendLine(GenerateComment("Generating junction tables"));
                        sb.AppendLine();
                        firstIteration = false;
                    }
                    sb.AppendLine(GenerateJunctionTable(foreignKey));
                }
            }

            return sb.ToString();
        }

		/// <summary>
		/// Generates the junction table. Uses composite primary key.
		/// </summary>
		/// <param name="foreignKey">The foreign key.</param>
		/// <returns></returns>
		private string GenerateJunctionTable(ForeignKeyConstraint foreignKey)
        {
            var sb = new StringBuilder();

            var tableName = string.Concat(foreignKey.PartOf.Name, foreignKey.References.PartOf.Name);
            sb.AppendLine(string.Format("CREATE TABLE {0}", ValidateString(tableName.Pluralize())));
            sb.AppendLine("(");

            //first foreign key
            sb.Append(string.Format("{0} {1} REFERENCES {2} ({3}),",
                ValidateString(foreignKey.PartOf.PrimaryKey.Name),
                SqlProperties.DataTypes[foreignKey.PartOf.PrimaryKey.DataType],
                ValidateString(foreignKey.PartOf.Name),
                ValidateString(foreignKey.PartOf.PrimaryKey.Name))
                .InsertTabsAtStartOfEachLine());
            //second foreign key
            sb.Append(string.Format("{0} {1} REFERENCES {2} ({3}),",
                ValidateString(foreignKey.References.Name),
                SqlProperties.DataTypes[foreignKey.References.DataType],
                ValidateString(foreignKey.References.PartOf.Name),
                ValidateString(foreignKey.References.Name))
                .InsertTabsAtStartOfEachLine());
            //composite primary key
            sb.Append(string.Format("PRIMARY KEY ({0}, {1})", ValidateString(foreignKey.PartOf.PrimaryKey.Name),
                ValidateString(foreignKey.References.Name)).InsertTabsAtStartOfEachLine());
            sb.AppendLine(");");

            return sb.ToString();
        }

		/// <summary>
		/// Generates the table.
		/// </summary>
		/// <param name="class">The class.</param>
		/// <returns></returns>
		private string GenerateTable(Class @class)
        {
            var sb = new StringBuilder();

            sb.AppendLine(string.Format("CREATE TABLE {0}", ValidateString(@class.Name)));
            sb.AppendLine("(");
            sb.Append(GenerateAttributes(@class).InsertTabsAtStartOfEachLine());
            sb.AppendLine(");");

            return sb.ToString();
        }

		/// <summary>
		/// Generates the attributes.
		/// </summary>
		/// <param name="class">The class.</param>
		/// <returns></returns>
		private string GenerateAttributes(Class @class)
        {
            var sb = new StringBuilder();

            if (@class.PrimaryKey != null)
            {
                sb.AppendLine(string.Format("{0} {1} PRIMARY KEY,", ValidateString(@class.PrimaryKey.Name),
                    SqlProperties.DataTypes[@class.PrimaryKey.DataType]));
            }

            foreach (var attribute in @class.Attributes)
            {
                sb.AppendLine(string.Format("{0} {1} {2},", 
					ValidateString(attribute.Name),
					SqlProperties.DataTypes[attribute.DataType],
					string.Empty));
            }

            foreach (var foreignKey in @class.Aggregations.Union(@class.Associations).Union(@class.Compositions))
            {
                if (foreignKey.TargetMultiplicity == MultiplicityType.One ||
                    foreignKey.TargetMultiplicity == MultiplicityType.ZeroToOne)
                {
                    sb.AppendLine(string.Format("{0} {1},", ValidateString(foreignKey.Name), SqlProperties.DataTypes[foreignKey.DataType]));
                }
            }

            foreach (var inheritedClass in @class.Children)
            {
                sb.AppendLine(string.Format("{0} {1},", ValidateString(inheritedClass.PrimaryKey.Name),
                    SqlProperties.DataTypes[inheritedClass.PrimaryKey.DataType]));
            }

            sb.Remove(sb.Length - 3, 3);

            return sb.ToString();
        }
    }
}
