﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using UmlDiagramDesigner.CodeGeneration.ConcreteGenerators;
using UmlDiagramDesigner.CodeGeneration.Contracts;
using UmlDiagramDesigner.CodeGeneration.Enums;
using UmlDiagramDesigner.Core.Enums;
using UmlDiagramDesigner.Data;

namespace UmlDiagramDesigner.CodeGeneration
{
    /// <summary>
    /// Base class for all Object-oriented language generators. Language has to support inheritance.
    /// </summary>
    public abstract class ObjectOrientedLanguageGenerator : GeneratorBase
    {

        protected ObjectOrientedLanguageGenerator(IOopLanguageProperties languageProperties) : base(languageProperties)
        {
            LanguageProperties = languageProperties;
        }

		/// <summary>
		/// Gets the required namespaces.
		/// </summary>
		/// <value>
		/// The required namespaces.
		/// </value>
		protected abstract IList<string> RequiredNamespaces { get; }

        public new IOopLanguageProperties LanguageProperties { get; private set; }

		/// <summary>
		/// Generates the class.
		/// </summary>
		/// <param name="class">The class.</param>
		/// <returns></returns>
		protected abstract string GenerateClass(Class @class);

		/// <summary>
		/// Generates the name of the class.
		/// </summary>
		/// <param name="class">The class.</param>
		/// <returns></returns>
		protected virtual string GenerateClassName(Class @class)
        {
            string result;
            string publicAccessibility;

            LanguageProperties.Accessibilities.TryGetValue(Accessibility.Public, out publicAccessibility);

            if (@class.Ancestor == null)
            {
                result = string.Format("{0} class {1}", publicAccessibility ?? string.Empty, @class.Name);
            }
            else
            {
                result = string.Format("{0} class {1} {2} {3}", publicAccessibility ?? string.Empty, @class.Name,
                    LanguageProperties.Features[LanguageFeature.Inheritance], @class.Ancestor.Name);
            }

            return result;
        }

		/// <summary>
		/// Generates the namespace imports.
		/// </summary>
		/// <returns></returns>
		protected virtual string GenerateNamespaceImports()
        {
            var sb = new StringBuilder();

            string keyword;
            if (LanguageProperties.Features.TryGetValue(LanguageFeature.NamespaceImport, out keyword))
            {
                string endLine;
                if (LanguageProperties.Features.TryGetValue(LanguageFeature.EndLineCharacter, out endLine))
                {
                    foreach (var ns in RequiredNamespaces)
                    {
                        sb.AppendLine(string.Format("{0} {1}{2}",keyword, ns, endLine));
                    }
                }
            }

            return sb.ToString();
        }

		/// <summary>
		/// Generates the namespace start (if not necessary, override empty).
		/// </summary>
		/// <param name="namespaceName">Name of the namespace.</param>
		/// <returns></returns>
		protected virtual string GenerateNamespaceStart(string namespaceName)
        {
            string namespaceString;
            string blockStart;
            if (LanguageProperties.Features.TryGetValue(LanguageFeature.Namespace, out namespaceString) &&
                LanguageProperties.Features.TryGetValue(LanguageFeature.BlockStart, out blockStart))
            {
                return string.Format("{0} {1}\n{2}",namespaceString, namespaceName,blockStart);
            }
            return string.Empty;
        }

		/// <summary>
		/// Generates the namespace end (if not necessary, override empty).
		/// </summary>
		/// <returns></returns>
		protected virtual string GenerateNamespaceEnd()
        {
            string blockEnd;
            if (LanguageProperties.Features.TryGetValue(LanguageFeature.BlockEnd, out blockEnd))
            {
                return blockEnd;
            }
            return string.Empty;
        }

		/// <summary>
		/// Generates the line end (will be inserted at the end of every line.
		/// </summary>
		/// <returns></returns>
		protected virtual string GenerateLineEnd()
        {
            string lineEnd;
            if (LanguageProperties.Features.TryGetValue(LanguageFeature.EndLineCharacter, out lineEnd))
            {
                return lineEnd;
            }
            return string.Empty;
        }

		/// <summary>
		/// Generates the start block (for example { in C#).
		/// </summary>
		/// <returns></returns>
		protected virtual string GenerateStartBlock()
        {
            string startBlock;
            if (LanguageProperties.Features.TryGetValue(LanguageFeature.BlockStart, out startBlock))
            {
                return startBlock;
            }
            return string.Empty;
        }

		/// <summary>
		/// Generates the end block (for example } in C#).
		/// </summary>
		/// <returns></returns>
		protected virtual string GenerateEndBlock()
        {
            string endBlock;
            if (LanguageProperties.Features.TryGetValue(LanguageFeature.BlockEnd, out endBlock))
            {
                return endBlock;
            }
            return string.Empty;
        }

		/// <summary>
		/// Determines the name of the navigation property name.
		/// Uses multiplicity.
		/// </summary>
		/// <param name="navigationProperty">The navigation property.</param>
		/// <returns></returns>
		/// <exception cref="System.ArgumentOutOfRangeException"></exception>
		protected virtual string DetermineNavPropertyName(ForeignKeyConstraint navigationProperty)
		{
			switch (navigationProperty.TargetMultiplicity)
			{
				case MultiplicityType.One:
				case MultiplicityType.ZeroToOne:
					return navigationProperty.Name;
				case MultiplicityType.ZeroToMany:
				case MultiplicityType.OneToMany:
					return navigationProperty.Name.Pluralize();
				default:
					throw new ArgumentOutOfRangeException();
			}
		}
	}
}
