﻿using System;
using System.Collections.Generic;

namespace UmlDiagramDesigner.Core
{
    /// <summary>
    /// Provides an ability for decoupled objects to exchange messages with each other.
    /// Implementation of Mediator pattern
    /// </summary>
    public static class Messenger
    {
        private static readonly Dictionary<Type, IList<ReceiverAction>> RegisteredRecipients;

        static Messenger()
        {
            RegisteredRecipients = new Dictionary<Type, IList<ReceiverAction>>();
        }

        /// <summary>
        /// Registers a generic action of T. All objects that register this type will receive appropriate data.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="receiver">The Receiver</param>
        /// <param name="action">The action.</param>
        /// <exception cref="Exception">A delegate callback throws an exception. </exception>
        public static void Register<T>(object receiver, Action<T> action)
        {
            IList<ReceiverAction> value;
            if (RegisteredRecipients.TryGetValue(typeof(T), out value))
            {
                value.Add(new ReceiverAction(receiver, o => action((T)o)));
            }
            else
            {
                var recipients = new List<ReceiverAction>() {new ReceiverAction(receiver, o => action((T) o))};
                RegisteredRecipients.Add(typeof(T), recipients);
            }
        }

        /// <summary>
        /// Unregisters the specified receiver.
        /// </summary>
        /// <typeparam name="T"></typeparam>
        /// <param name="receiver">The receiver.</param>
        /// <param name="action">The action.</param>
        public static void Unregister<T>(object receiver, Action<T> action)
        {
            IList<ReceiverAction> value;
            if (RegisteredRecipients.TryGetValue(typeof(T), out value))
            {
                for (int i = 0; i < value.Count; i++)
                {
					if(value[i].Receiver == receiver)
						value.RemoveAt(i);
                }
            }
        }

        /// <summary>
        /// Sends the specified message too all registered recipients.
        /// </summary>
        /// <param name="message">The message data.</param>
        /// <exception cref="Exception">A delegate callback throws an exception. </exception>
        public static void Send<T>(T message)
        {
            IList<ReceiverAction> value;
            if (RegisteredRecipients.TryGetValue(typeof(T), out value))
            {
                foreach (var receiverAction in value)
                {
                    receiverAction.Action.Invoke(message);
                }
            }
        }

		/// <summary>
		/// Internal class for coupling message with specific reciever.
		/// Used for unregistering from <see cref="Messenger"/>
		/// </summary>
		private class ReceiverAction
        {
            public ReceiverAction(object receiver, Action<object> action)
            {
                Receiver = receiver;
                Action = action;
            }
            public object Receiver { get; set; }
            public Action<object> Action { get; set; }
        }
    }
}