﻿using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using UmlDiagramDesigner.Core.Enums;
using UmlDiagramDesigner.Data;
using UmlDiagramDesigner.DTO;
using UmlDiagramDesigner.Models;
using UmlDiagramDesigner.Test.Mocks;
using UmlDiagramDesigner.ViewModels;
using UmlDiagramDesigner.ViewModels.DependentElements;
using UmlDiagramDesigner.ViewModels.DiagramItems;
using UmlDiagramDesigner.ViewModels.DiagramRelationships;

namespace UmlDiagramDesigner.Test
{
    [TestClass]
    public class DiagramConversionTest
    {
        private readonly DiagramToDataModelConverter _converter = new DiagramToDataModelConverter();


        [TestMethod]
        public void Convert_Empty_Diagram()
        {
            var canvas = new CanvasViewModel()
            {
                Name = null,
                Items = new ObservableCollection<DiagramEntityViewModelBase>()
            };

            var result = _converter.Convert(canvas);

            Assert.IsNotNull(result);
            Assert.AreEqual(DiagramToDataModelConverter.DefaultDiagramName,result.Name);
            Assert.AreEqual(0,result.Classes.Count);
            Assert.AreEqual(0, result.Views.Count);
        }

        [TestMethod]
        public void Convert_Diagram_Name()
        {
            var canvas = new CanvasViewModel()
            {
                Name = "TestName",
            };
            var result = _converter.Convert(canvas);

            Assert.IsNotNull(result);
            Assert.AreEqual("TestName", result.Name);
        }
        
        [TestMethod]
        public void Convert_Single_Class()
        {
            var canvas = new CanvasViewModel()
            {
                Name = null,
                Items = new ObservableCollection<DiagramEntityViewModelBase>()
            };
            var entity = new ClassDiagramItemViewModel()
            {
                Name = "TestClass",
                Attributes = new ObservableCollection<AttributeViewModel>()
                {
                    new AttributeViewModel() {Name = "Name", DataType = DbDataType.Varchar, IsNullable = true},
                    new AttributeViewModel() {Name = "Age", DataType = DbDataType.Integer, IsNotNull = true},
                    new AttributeViewModel() {Name = "Id", DataType = DbDataType.Integer, IsPrimaryKey = true}
                }
            };
            canvas.Items.Add(entity);

            var result = _converter.Convert(canvas);

            //basic assertions
            Assert.IsNotNull(result);
            Assert.AreEqual("TestClass",result.Classes.First().Name);
            Assert.AreEqual(1,result.Classes.Count);
            Assert.AreEqual(0, result.Views.Count);
            //concrete assertions
            Assert.AreEqual(2,result.Classes.First().Attributes.Count); //primary key is should not be in attributes
            Assert.IsNotNull(result.Classes.First().PrimaryKey);
            Assert.AreEqual(typeof (NullableConstraint),
                result.Classes.First().Attributes.First().Constraints.First().GetType());
            Assert.AreEqual(typeof (NotNullConstraint),
                result.Classes.First().Attributes.Skip(1).First().Constraints.First().GetType());
            Assert.AreEqual(entity.Attributes[0].DataType,result.Classes.First().Attributes.First().DataType);
            Assert.AreEqual(entity.Attributes[1].DataType, result.Classes.First().Attributes.Skip(1).First().DataType);
        }

        [TestMethod]
        public void Convert_Single_View()
        {
            var canvas = new CanvasViewModel()
            {
                Name = null,
                Items = new ObservableCollection<DiagramEntityViewModelBase>()
            };
            var entity = new ClassDiagramItemViewModel()
            {
                Name = "TestClass",
                Attributes = new ObservableCollection<AttributeViewModel>()
                {
                    new AttributeViewModel() {Name = "Name", DataType = DbDataType.Varchar, IsNullable = true},
                    new AttributeViewModel() {Name = "Age", DataType = DbDataType.Integer, IsNotNull = true},
                    new AttributeViewModel() {Name = "Id", DataType = DbDataType.Integer, IsPrimaryKey = true}
                }
            };
            var view = new ViewDiagramItemViewModel()
            {
                Name = "TestView",
                Attributes = new ObservableCollection<AttributeBySourceDTO>()
                {
                    new AttributeBySourceDTO() {Attribute = entity.Attributes[0], Source = entity, IsSelected = true},
                    new AttributeBySourceDTO() {Attribute = entity.Attributes[1], Source = entity, IsSelected = true},
                    new AttributeBySourceDTO() {Attribute = entity.Attributes[2], Source = entity}
                }
            };
            canvas.Items.Add(entity);
            canvas.Items.Add(view);

            var result = _converter.Convert(canvas);

            Assert.IsNotNull(result);
            Assert.AreEqual(1,result.Classes.Count);
            Assert.AreEqual(1, result.Views.Count);
            Assert.AreEqual("TestView",result.Views.First().Name);
            Assert.AreEqual(2,result.Views.First().Attributes.Count); //third attribute has IsSelected false, so it shouldn't be included
            Assert.AreEqual(entity.Attributes[0].DataType,result.Views.First().Attributes.First().DataType);
         }

        [TestMethod]
        public void Convert_Three_Inherited_Classes()
        {
            var canvas = new CanvasViewModel()
            {
                Name = null,
                Items = new ObservableCollection<DiagramEntityViewModelBase>()
            };
            var first = new ClassDiagramItemViewModel(){Name = "First"};
            var second = new ClassDiagramItemViewModel() {Name = "Second"};
            var third = new ClassDiagramItemViewModel() {Name = "Third"};

            var firstInheritance = new InheritanceViewModel { Source = second, Target = first };//second -> first
            var secondInheritance = new InheritanceViewModel { Source = third, Target = second };//third -> second

            first.Relationships.Add(firstInheritance);
            second.Relationships.Add(firstInheritance);
            second.Relationships.Add(secondInheritance);
            third.Relationships.Add(secondInheritance);

            canvas.Items.Add(first);
            canvas.Items.Add(second);
            canvas.Items.Add(third);

            var result = _converter.Convert(canvas);

            Assert.IsNotNull(result);
            Assert.AreEqual(3,result.Classes.Count);
            Assert.AreSame(result.Classes.First(),result.Classes.Skip(1).First().Ancestor);
            Assert.AreSame(result.Classes.Skip(1).First(), result.Classes.Skip(2).First().Ancestor);
        }

        [TestMethod]
        public void Convert_Aggregation_Relationship()
        {
            var canvas = new CanvasViewModel()
            {
                Name = null,
                Items = new ObservableCollection<DiagramEntityViewModelBase>()
            };

            var first = new ClassDiagramItemViewModel {Name = "First"};
            var second = new ClassDiagramItemViewModel { Name = "Second" };

            var relationship = new AggregationViewModel {Source = first, Target = second};

            first.Relationships.Add(relationship);
            second.Relationships.Add(relationship);
            first.Attributes.Add(new AttributeViewModel {IsForeignKey = true,Name = "ForeignKey", DataType = DbDataType.Integer});
            second.Attributes.Add(new AttributeViewModel { IsPrimaryKey= true, Name = "PrimaryKey", DataType = DbDataType.Integer});

            canvas.Items.Add(first);
            canvas.Items.Add(second);
            canvas.Items.Add(relationship);

            var result = _converter.Convert(canvas);

            Assert.IsNotNull(result);
            Assert.AreEqual(2,result.Classes.Count);
            Assert.AreEqual(1,result.Classes.First().Aggregations.Count);
            Assert.AreEqual(result.Classes.First().Aggregations.First().References.PartOf,result.Classes.First());
        }

        [TestMethod]
        public void Convert_Composition_Relationship()
        {
            var canvas = new CanvasViewModel()
            {
                Name = null,
                Items = new ObservableCollection<DiagramEntityViewModelBase>()
            };

            var first = new ClassDiagramItemViewModel { Name = "First" };
            var second = new ClassDiagramItemViewModel { Name = "Second" };

            var relationship = new CompositionViewModel { Source = first, Target = second };

            first.Relationships.Add(relationship);
            second.Relationships.Add(relationship);
            first.Attributes.Add(new AttributeViewModel { IsForeignKey = true, Name = "ForeignKey", DataType = DbDataType.Integer });
            second.Attributes.Add(new AttributeViewModel { IsPrimaryKey = true, Name = "PrimaryKey", DataType = DbDataType.Integer });

            canvas.Items.Add(first);
            canvas.Items.Add(second);
            canvas.Items.Add(relationship);

            var result = _converter.Convert(canvas);

            Assert.IsNotNull(result);
            Assert.AreEqual(2, result.Classes.Count);
            Assert.AreEqual(1, result.Classes.First().Compositions.Count);
            Assert.AreEqual(result.Classes.First().Compositions.First().References.PartOf, result.Classes.First());
        }

        [TestMethod]
        public void Convert_Association_Relationship()
        {
            var canvas = new CanvasViewModel
            {
                Name = null,
                Items = new ObservableCollection<DiagramEntityViewModelBase>()
            };

            var first = new ClassDiagramItemViewModel { Name = "First" };
            var second = new ClassDiagramItemViewModel { Name = "Second" };

            var relationship = new AssociationViewModel { Source = first, Target = second };

            first.Relationships.Add(relationship);
            second.Relationships.Add(relationship);
            first.Attributes.Add(new AttributeViewModel { IsForeignKey = true, Name = "ForeignKey", DataType = DbDataType.Integer });
            second.Attributes.Add(new AttributeViewModel { IsPrimaryKey = true, Name = "PrimaryKey", DataType = DbDataType.Integer });

            relationship.SourceForeignKey = first.Attributes[0];

            canvas.Items.Add(first);
            canvas.Items.Add(second);
            canvas.Items.Add(relationship);

            var result = _converter.Convert(canvas);

            Assert.IsNotNull(result);
            Assert.AreEqual(2, result.Classes.Count);
            Assert.AreEqual(1, result.Classes.First().Associations.Count);
            Assert.AreEqual(result.Classes.First().Associations.First().References.PartOf, result.Classes.First());
        }
    }
}

