﻿using System;
using System.Collections.Generic;
using System.Linq;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using UmlDiagramDesigner.Core;
using UmlDiagramDesigner.Core.Graph;
using UmlDiagramDesigner.Test.Mocks;
using UmlDiagramDesigner.ViewModels.DiagramRelationships;

namespace UmlDiagramDesigner.Test
{
    [TestClass]
    public class TarjansAlgorithmTest
    {
        private readonly ICycleDetector _tarjan = new TarjansAlgorithm();

        /// <summary>
        /// No vertices
        /// </summary>
        [TestMethod]
        public void NoVertex_NoConnection_NoCycle()
        {
            IIndexedGraph graph = new IndexedGraphMock { Vertices = new List<IIndexedVertex>() };

            var cycles = _tarjan.DetectCycles(graph);

            Assert.IsNotNull(cycles, "Empty non-null collection should have been returned.");
            Assert.AreEqual(0, cycles.Count, "There should be no cycles in graph with no vertices.");
        }

        /// <summary>
        /// A
        /// </summary>
        [TestMethod]
        public void SingleVertex_NoConnection_NoCycle()
        {
            IIndexedVertex vertex = new IndexedVertexMock();
            IIndexedGraph graph = new IndexedGraphMock(){Vertices = new List<IIndexedVertex>{vertex}};

            var cycles = _tarjan.DetectCycles(graph);

            Assert.IsNotNull(cycles, "Empty non-null collection should have been returned.");
            Assert.AreEqual(0,cycles.Count,"There should be no cycles in graph with a single node");
        }

        /// <summary>
        /// A | B
        /// </summary>
        [TestMethod]
        public void TwoVertices_NoConnection_NoCycle()
        {
            IIndexedVertex vertexA = new IndexedVertexMock();
            IIndexedVertex vertexB = new IndexedVertexMock();            

            IIndexedGraph graph = new IndexedGraphMock
            { Vertices = new List<IIndexedVertex> { vertexA, vertexB } };

            var cycles = _tarjan.DetectCycles(graph);

            Assert.IsNotNull(cycles, "Empty non-null collection should have been returned.");
            Assert.AreEqual(0, cycles.Count, "There should be no cycles in graph with two independent vertices.");
        }

        /// <summary>
        /// A -> B
        /// </summary>
        [TestMethod]
        public void TwoVertices_SingleConnection_NoCycle()
        {
            var vertexA = new IndexedVertexMock();
            var vertexB = new IndexedVertexMock();

            vertexA.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexB,new InheritanceRelationshipMock()));

            IIndexedGraph graph = new IndexedGraphMock { Vertices = new List<IIndexedVertex> { vertexA, vertexB } };

            var cycles = _tarjan.DetectCycles(graph);

            Assert.IsNotNull(cycles, "Empty non-null collection should have been returned.");
            Assert.AreEqual(0, cycles.Count, "There should be no cycles in graph with two independent vertices.");
        }

        /// <summary>
        /// A -> B -> A
        /// </summary>
        [TestMethod]
        public void TwoVertices_TwoConnections_OneCycle()
        {
            var vertexA = new IndexedVertexMock();
            var vertexB = new IndexedVertexMock();

            vertexA.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexB, new InheritanceRelationshipMock()));
            vertexB.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexA, new InheritanceRelationshipMock()));

            IIndexedGraph graph = new IndexedGraphMock { Vertices = new List<IIndexedVertex> { vertexA, vertexB } };

            var cycles = _tarjan.DetectCycles(graph);

            Assert.IsNotNull(cycles, "Empty non-null collection should have been returned.");
            Assert.AreEqual(1, cycles.Count, "There should be a single cycle between vertexA and vertexB.");
            Assert.IsTrue(cycles.First().Contains(vertexA), "Cycle should contain vertexA");
            Assert.IsTrue(cycles.First().Contains(vertexB), "Cycle should contain vertexB");
        }

        /// <summary>
        /// A -> B --> A
        /// </summary>
        [TestMethod]
        public void TwoVertices_TwoDifferentConnections_SingleCycle()
        {
            var vertexA = new IndexedVertexMock();
            var vertexB = new IndexedVertexMock();

            vertexA.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexB, new InheritanceRelationshipMock()));
            vertexB.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexA, new CompositionRelationshipMock()));

            IIndexedGraph graph = new IndexedGraphMock { Vertices = new List<IIndexedVertex> { vertexA, vertexB } };

            var cycles = _tarjan.DetectCycles(graph);

            Assert.IsNotNull(cycles, "Empty non-null collection should have been returned.");
            Assert.AreEqual(1, cycles.Count, "There should be a single cycle between vertexA and vertexB.");
            Assert.IsTrue(cycles.First().Contains(vertexA), "Cycle should contain vertexA");
            Assert.IsTrue(cycles.First().Contains(vertexB), "Cycle should contain vertexB");
        }

        /// <summary>
        /// A -> B --> A
        /// </summary>
        [TestMethod]
        public void TwoVertices_TwoDifferentConnections_NoCycle()
        {
            var vertexA = new IndexedVertexMock();
            var vertexB = new IndexedVertexMock();

            vertexA.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexB, new InheritanceRelationshipMock()));
            vertexB.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexA, new CompositionRelationshipMock()));

            IIndexedGraph graph = new IndexedGraphMock { Vertices = new List<IIndexedVertex> { vertexA, vertexB } };

            var cycles = _tarjan.DetectCycles<InheritanceRelationshipMock>(graph);

            Assert.IsNotNull(cycles, "Empty non-null collection should have been returned.");
            Assert.AreEqual(0, cycles.Count, "There shouldnt be any cycles between vertexA and vertexB because of different type.");
        }


        /// <summary>
        /// A -> B | C
        /// </summary>
        [TestMethod]
        public void ThreeVertices_TwoConnections_NoCycle()
        {
            var vertexA = new IndexedVertexMock();
            var vertexB = new IndexedVertexMock();
            var vertexC = new IndexedVertexMock();

            vertexA.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexB, new InheritanceRelationshipMock()));

            IIndexedGraph graph = new IndexedGraphMock 
            { Vertices = new List<IIndexedVertex> { vertexA, vertexB, vertexC } };

            var cycles = _tarjan.DetectCycles<InheritanceRelationshipMock>(graph);

            Assert.IsNotNull(cycles, "Empty non-null collection should have been returned.");
            Assert.AreEqual(0, cycles.Count, "There should not be any cycles.");
        }

        /// <summary>
        /// A -> B -> C -> A
        /// </summary>
        [TestMethod]
        public void ThreeVertices_ThreeConnections_SingleCycle()
        {
            var vertexA = new IndexedVertexMock();
            var vertexB = new IndexedVertexMock();
            var vertexC = new IndexedVertexMock();

            vertexA.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexB, new InheritanceRelationshipMock()));
            vertexB.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexC, new InheritanceRelationshipMock()));
            vertexC.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexA, new InheritanceRelationshipMock()));

            IIndexedGraph graph = new IndexedGraphMock { Vertices = new List<IIndexedVertex> { vertexA, vertexB, vertexC } };

            var cycles = _tarjan.DetectCycles<InheritanceRelationshipMock>(graph);

            Assert.IsNotNull(cycles, "Empty non-null collection should have been returned.");
            Assert.AreEqual(1, cycles.Count, "There should be a single cycle.");
            Assert.IsTrue(cycles.First().Contains(vertexA), "Cycle should contain vertexA");
            Assert.IsTrue(cycles.First().Contains(vertexB), "Cycle should contain vertexB");
            Assert.IsTrue(cycles.First().Contains(vertexC), "Cycle should contain vertexC");
        }

        /// <summary>
        /// A -> B -> C -> A | D -> E -> F -> D
        /// </summary>
        [TestMethod]
        public void SixVertices_SixConnections_TwoCycles()
        {
            var vertexA = new IndexedVertexMock();
            var vertexB = new IndexedVertexMock();
            var vertexC = new IndexedVertexMock();

            vertexA.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexB, new InheritanceRelationshipMock()));
            vertexB.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexC, new InheritanceRelationshipMock()));
            vertexC.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexA, new InheritanceRelationshipMock()));

            var vertexD = new IndexedVertexMock();
            var vertexE = new IndexedVertexMock();
            var vertexF = new IndexedVertexMock();

            vertexD.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexE, new InheritanceRelationshipMock()));
            vertexE.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexF, new InheritanceRelationshipMock()));
            vertexF.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexD, new InheritanceRelationshipMock()));

            IIndexedGraph graph = new IndexedGraphMock
            {Vertices = new List<IIndexedVertex> {vertexA, vertexB, vertexC, vertexD, vertexE, vertexF}};

            var cycles = _tarjan.DetectCycles<InheritanceRelationshipMock>(graph);

            Assert.IsNotNull(cycles, "Empty non-null collection should have been returned.");
            Assert.AreEqual(2, cycles.Count, "There should be two cycles.");

            Assert.IsTrue(cycles.First().Contains(vertexA), "First cycle should contain vertexA");
            Assert.IsTrue(cycles.First().Contains(vertexB), "First cycle should contain vertexB");
            Assert.IsTrue(cycles.First().Contains(vertexC), "First cycle should contain vertexC");

            Assert.IsTrue(cycles.Skip(1).First().Contains(vertexD), "Second cycle should contain vertexD");
            Assert.IsTrue(cycles.Skip(1).First().Contains(vertexE), "Second cycle should contain vertexE");
            Assert.IsTrue(cycles.Skip(1).First().Contains(vertexF), "Second cycle should contain vertexF");
        }

        /// <summary>
        /// A -> B -> C -> A | D -> E -> F -> D
        /// </summary>
        [TestMethod]
        public void SixVertices_SixConnections_OneCycle()
        {
            var vertexA = new IndexedVertexMock();
            var vertexB = new IndexedVertexMock();
            var vertexC = new IndexedVertexMock();

            vertexA.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexB, new InheritanceRelationshipMock()));
            vertexB.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexC, new InheritanceRelationshipMock()));
            vertexC.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexA, new InheritanceRelationshipMock()));

            var vertexD = new IndexedVertexMock();
            var vertexE = new IndexedVertexMock();
            var vertexF = new IndexedVertexMock();

            vertexD.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexE, new CompositionRelationshipMock()));
            vertexE.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexF, new CompositionRelationshipMock()));
            vertexF.Neighbours.Add(new Tuple<IIndexedVertex, RelationshipViewModelBase>(vertexD, new CompositionRelationshipMock()));

            IIndexedGraph graph = new IndexedGraphMock { Vertices = new List<IIndexedVertex> { vertexA, vertexB, vertexC, vertexD, vertexE, vertexF } };

            var cycles = _tarjan.DetectCycles<InheritanceRelationshipMock>(graph);

            Assert.IsNotNull(cycles, "Empty non-null collection should have been returned.");
            Assert.AreEqual(1, cycles.Count, "There should be one cycle.");

            Assert.IsTrue(cycles.First().Contains(vertexA), "First cycle should contain vertexA");
            Assert.IsTrue(cycles.First().Contains(vertexB), "First cycle should contain vertexB");
            Assert.IsTrue(cycles.First().Contains(vertexC), "First cycle should contain vertexC");
        }
    }
}


        
        
        
        
        

        
        

        

        

        
        
        
        
        

        
        
        
        
        

        
        

        

        

        
        
        


        
        
        
        
        
        

        

        
        

        

        
        
        